'''
fs.contrib.tahoelafs
====================

This modules provides a PyFilesystem interface to the Tahoe Least Authority
File System. Tahoe-LAFS is a distributed, encrypted, fault-tolerant storage
system:

    http://tahoe-lafs.org/

You will need access to a Tahoe-LAFS "web api" service.

Example (it will use publicly available (but slow) Tahoe-LAFS cloud)::

    from fs.contrib.tahoelafs import TahoeLAFS, Connection
    dircap = TahoeLAFS.createdircap(webapi='http://insecure.tahoe-lafs.org')
    print "Your dircap (unique key to your storage directory) is", dircap
    print "Keep it safe!"
    fs = TahoeLAFS(dircap, autorun=False, webapi='http://insecure.tahoe-lafs.org')
    f = fs.open("foo.txt", "a")
    f.write('bar!')
    f.close()
    print "Now visit %s and enjoy :-)" % fs.getpathurl('foo.txt')

When any problem occurred, you can turn on internal debugging messages::

    import logging    
    l = logging.getLogger()
    l.setLevel(logging.DEBUG)
    l.addHandler(logging.StreamHandler(sys.stdout))

    ... your Python code using TahoeLAFS ...
    
TODO:

   * unicode support
   * try network errors / bad happiness
   * exceptions
   * tests    
   * sanitize all path types (., /)
   * support for extra large file uploads (poster module)
   * Possibility to block write until upload done (Tahoe mailing list)
   * Report something sane when Tahoe crashed/unavailable
   * solve failed unit tests (makedir_winner, ...)
   * file times
   * docs & author
   * python3 support
   * remove creating blank files (depends on FileUploadManager)
   
TODO (Not TahoeLAFS specific tasks):
   * RemoteFileBuffer on the fly buffering support
   * RemoteFileBuffer unit tests
   * RemoteFileBuffer submit to trunk
   * Implement FileUploadManager + faking isfile/exists of just processing file
   * pyfilesystem docs is outdated (rename, movedir, ...)  

'''


import stat as statinfo

import logging
from logging import DEBUG, INFO, ERROR, CRITICAL

import fs
import fs.errors as errors
from fs.path import abspath, relpath, normpath, dirname, pathjoin
from fs.base import FS, NullFile
from fs import _thread_synchronize_default, SEEK_END
from fs.remote import CacheFSMixin, RemoteFileBuffer
from fs.base import fnmatch, NoDefaultMeta

from util import TahoeUtil
from connection import Connection   

from six import b

logger = fs.getLogger('fs.tahoelafs')

def _fix_path(func):
    """Method decorator for automatically normalising paths."""
    def wrapper(self, *args, **kwds):
        if len(args):
            args = list(args)
            args[0] = _fixpath(args[0])
        return func(self, *args, **kwds)
    return wrapper


def _fixpath(path):
    """Normalize the given path."""
    return abspath(normpath(path))
    
     

class _TahoeLAFS(FS):
    """FS providing raw access to a Tahoe-LAFS Filesystem.

    This class implements all the details of interacting with a Tahoe-backed
    filesystem, but you probably don't want to use it in practice.  Use the
    TahoeLAFS class instead, which has some internal caching to improve
    performance.
    """
    
    _meta = { 'virtual' : False,
              'read_only' : False,
              'unicode_paths' : True,
              'case_insensitive_paths' : False,
              'network' : True
             }
        

    def __init__(self, dircap, largefilesize=10*1024*1024, webapi='http://127.0.0.1:3456'):
        '''Creates instance of TahoeLAFS.
            
            :param dircap: special hash allowing user to work with TahoeLAFS directory.
            :param largefilesize: - Create placeholder file for files larger than this treshold.
                Uploading and processing of large files can last extremely long (many hours),
                so placing this placeholder can help you to remember that upload is processing.
                Setting this to None will skip creating placeholder files for any uploads.
        '''
        self.dircap = dircap if not dircap.endswith('/') else dircap[:-1]
        self.largefilesize = largefilesize
        self.connection = Connection(webapi)
        self.tahoeutil = TahoeUtil(webapi)
        super(_TahoeLAFS, self).__init__(thread_synchronize=_thread_synchronize_default)       
        
    def __str__(self):
        return "<TahoeLAFS: %s>" % self.dircap 
    
    @classmethod
    def createdircap(cls, webapi='http://127.0.0.1:3456'):
        return TahoeUtil(webapi).createdircap()

    def getmeta(self,meta_name,default=NoDefaultMeta):
        if meta_name == "read_only":
            return self.dircap.startswith('URI:DIR2-RO')
        return super(_TahoeLAFS,self).getmeta(meta_name,default)
    
    @_fix_path
    def open(self, path, mode='r', **kwargs):
        self._log(INFO, 'Opening file %s in mode %s' % (path, mode))        
        newfile = False
        if not self.exists(path):
            if 'w' in mode or 'a' in mode:
                newfile = True
            else:
                self._log(DEBUG, "File %s not found while opening for reads" % path)
                raise errors.ResourceNotFoundError(path)
        elif self.isdir(path):
            self._log(DEBUG, "Path %s is directory, not a file" % path)
            raise errors.ResourceInvalidError(path)
        elif 'w' in mode:
            newfile = True
        
        if newfile:
            self._log(DEBUG, 'Creating empty file %s' % path)
            if self.getmeta("read_only"):
                raise errors.UnsupportedError('read only filesystem')
            self.setcontents(path, b(''))
            handler = NullFile()
        else:
            self._log(DEBUG, 'Opening existing file %s for reading' % path)
            handler = self.getrange(path,0)
        
        return RemoteFileBuffer(self, path, mode, handler,
                    write_on_flush=False)

    @_fix_path
    def desc(self, path):
        try:
            return self.getinfo(path)
        except:
            return ''
    
    @_fix_path
    def exists(self, path):
        try:
            self.getinfo(path)
            self._log(DEBUG, "Path %s exists" % path)
            return True
        except errors.ResourceNotFoundError:
            self._log(DEBUG, "Path %s does not exists" % path)
            return False
        except errors.ResourceInvalidError:
            self._log(DEBUG, "Path %s does not exists, probably misspelled URI" % path)
            return False
     
    @_fix_path
    def getsize(self, path):
        try:
            size = self.getinfo(path)['size']
            self._log(DEBUG, "Size of %s is %d" % (path, size))
            return size
        except errors.ResourceNotFoundError:
            return 0
    
    @_fix_path
    def isfile(self, path):
        try:
            isfile = (self.getinfo(path)['type'] == 'filenode')
        except errors.ResourceNotFoundError:
            #isfile = not path.endswith('/')
            isfile = False
        self._log(DEBUG, "Path %s is file: %d" % (path, isfile))
        return isfile
    
    @_fix_path        
    def isdir(self, path):
        try:
            isdir = (self.getinfo(path)['type'] == 'dirnode')
        except errors.ResourceNotFoundError:
            isdir = False
        self._log(DEBUG, "Path %s is directory: %d" % (path, isdir))
        return isdir

    
    def listdir(self, *args, **kwargs):
        return [ item[0] for item in self.listdirinfo(*args, **kwargs) ]        

    def listdirinfo(self, *args, **kwds):
        return list(self.ilistdirinfo(*args,**kwds))

    def ilistdir(self, *args, **kwds):
        for item in self.ilistdirinfo(*args,**kwds):
            yield item[0]
    
    @_fix_path
    def ilistdirinfo(self, path="/", wildcard=None, full=False, absolute=False,
                    dirs_only=False, files_only=False):
        self._log(DEBUG, "Listing directory (listdirinfo) %s" % path)
        
        if dirs_only and files_only:
            raise ValueError("dirs_only and files_only can not both be True")
        
        for item in self.tahoeutil.list(self.dircap, path):
            if dirs_only and item['type'] == 'filenode':
                continue
            elif files_only and item['type'] == 'dirnode':
                continue
            
            if wildcard is not None:
                if isinstance(wildcard,basestring):
                    if not fnmatch.fnmatch(item['name'], wildcard):
                        continue
                else:
                    if not wildcard(item['name']):
                        continue
            
            if full:
                item_path = relpath(pathjoin(path, item['name']))
            elif absolute:
                item_path = abspath(pathjoin(path, item['name']))    
            else:
                item_path = item['name']
            
            yield (item_path, item)
     
    @_fix_path
    def remove(self, path):
        self._log(INFO, 'Removing file %s' % path)
        if self.getmeta("read_only"):
            raise errors.UnsupportedError('read only filesystem')

        if not self.isfile(path):
            if not self.isdir(path):
                raise errors.ResourceNotFoundError(path)
            raise errors.ResourceInvalidError(path)
        
        try:
            self.tahoeutil.unlink(self.dircap, path)
        except Exception, e:
            raise errors.ResourceInvalidError(path)
    
    @_fix_path
    def removedir(self, path, recursive=False, force=False):
        self._log(INFO, "Removing directory %s" % path) 
        if self.getmeta("read_only"):
            raise errors.UnsupportedError('read only filesystem')
        if not self.isdir(path):
            if not self.isfile(path):
                raise errors.ResourceNotFoundError(path)
            raise errors.ResourceInvalidError(path)
        if not force and self.listdir(path):
            raise errors.DirectoryNotEmptyError(path)
        
        self.tahoeutil.unlink(self.dircap, path)

        if recursive and path != '/':
            try:
                self.removedir(dirname(path), recursive=True)
            except errors.DirectoryNotEmptyError:
                pass
    
    @_fix_path
    def makedir(self, path, recursive=False, allow_recreate=False):
        self._log(INFO, "Creating directory %s" % path)
        if self.getmeta("read_only"):
            raise errors.UnsupportedError('read only filesystem')       
        if self.exists(path):
            if not self.isdir(path):
                raise errors.ResourceInvalidError(path)
            if not allow_recreate: 
                raise errors.DestinationExistsError(path)
        if not recursive and not self.exists(dirname(path)):
            raise errors.ParentDirectoryMissingError(path)
        self.tahoeutil.mkdir(self.dircap, path)
        
    def movedir(self, src, dst, overwrite=False):
        self.move(src, dst, overwrite=overwrite)
    
    def move(self, src, dst, overwrite=False):
        self._log(INFO, "Moving file from %s to %s" % (src, dst))
        if self.getmeta("read_only"):
            raise errors.UnsupportedError('read only filesystem')
        src = _fixpath(src)
        dst = _fixpath(dst)
        if not self.exists(dirname(dst)):
            raise errors.ParentDirectoryMissingError(dst)
        if not overwrite and self.exists(dst):
            raise errors.DestinationExistsError(dst)
        self.tahoeutil.move(self.dircap, src, dst)

    def rename(self, src, dst):
        self.move(src, dst)
        
    def copy(self, src, dst, overwrite=False, chunk_size=16384):
        if self.getmeta("read_only"):
            raise errors.UnsupportedError('read only filesystem')
        # FIXME: this is out of date; how to do native tahoe copy?
        # FIXME: Workaround because isfile() not exists on _TahoeLAFS
        FS.copy(self, src, dst, overwrite, chunk_size)
        
    def copydir(self, src, dst, overwrite=False, ignore_errors=False, chunk_size=16384):
        if self.getmeta("read_only"):
            raise errors.UnsupportedError('read only filesystem')
        # FIXME: this is out of date; how to do native tahoe copy?
        # FIXME: Workaround because isfile() not exists on _TahoeLAFS
        FS.copydir(self, src, dst, overwrite, ignore_errors, chunk_size)
       
    
    def _log(self, level, message):
        if not logger.isEnabledFor(level): return
        logger.log(level, u'(%d) %s' % (id(self),
                                unicode(message).encode('ASCII', 'replace')))
        
    @_fix_path
    def getpathurl(self, path, allow_none=False, webapi=None):
        '''
            Retrieve URL where the file/directory is stored
        '''
        if webapi == None:
            webapi = self.connection.webapi
        self._log(DEBUG, "Retrieving URL for %s over %s" % (path, webapi))
        path = self.tahoeutil.fixwinpath(path, False)
        return u"%s/uri/%s%s" % (webapi, self.dircap, path)

    @_fix_path
    def getrange(self, path, offset, length=None):
        return self.connection.get(u'/uri/%s%s' % (self.dircap, path),
                    offset=offset, length=length)
       
    @_fix_path             
    def setcontents(self, path, file, chunk_size=64*1024):    
        self._log(INFO, 'Uploading file %s' % path)
        size=None
        
        if self.getmeta("read_only"):
            raise errors.UnsupportedError('read only filesystem')
        
        # Workaround for large files:
        # First create zero file placeholder, then
        # upload final content.
        if self.largefilesize != None and getattr(file, 'read', None):
            # As 'file' can be also a string, need to check,
            # if 'file' looks like duck. Sorry, file.
            file.seek(0, SEEK_END)
            size = file.tell()
            file.seek(0)

            if size > self.largefilesize:
                self.connection.put(u'/uri/%s%s' % (self.dircap, path),
                    "PyFilesystem.TahoeLAFS: Upload started, final size %d" % size)

        self.connection.put(u'/uri/%s%s' % (self.dircap, path), file, size=size)

    @_fix_path
    def getinfo(self, path): 
        self._log(INFO, 'Reading meta for %s' % path)
        info = self.tahoeutil.info(self.dircap, path)        
        #import datetime
        #info['created_time'] = datetime.datetime.now()
        #info['modified_time'] = datetime.datetime.now()
        #info['accessed_time'] = datetime.datetime.now()
        if info['type'] == 'filenode':
            info["st_mode"] = 0x700 | statinfo.S_IFREG
        elif info['type'] == 'dirnode':
            info["st_mode"] = 0x700 | statinfo.S_IFDIR
        return info



class TahoeLAFS(CacheFSMixin,_TahoeLAFS):
    """FS providing cached access to a Tahoe Filesystem.

    This class is the preferred means to access a Tahoe filesystem.  It
    maintains an internal cache of recently-accessed metadata to speed
    up operations.
    """

    def __init__(self, *args, **kwds):
        kwds.setdefault("cache_timeout",60)
        super(TahoeLAFS,self).__init__(*args,**kwds)





# All built-in recognizer plugins should be listed in this file so that the
# application can import this single file and determine the default plugins.
# In addition, a list of plugins is expected so that the framework can import
# all built-in plugins at once.
#
# A cog script is included to automatically generate the expected code. Test with::
#
#    cog.py __init__.py
#
# which prints to standard output. To place the generated code in the file, use::
#
#    cog.py -r __init__.py
#
# Note that the cog script works when called from the top directory (.i.e.  at
# the same level as the omnivore directory) or this directory.

# [[[cog
#import os
#import sys
#import inspect
#import imp
#
#from envisage.api import Plugin
#
#cwd = os.getcwd()
#cog.msg("working dir : %s" % cwd)
#path = os.path.dirname(os.path.join(cwd, cog.inFile))
#cog.msg("scanning dir: %s" % path)
#top = os.path.abspath(os.path.join(path, "../../..")) # so absolute imports of omnivore will work
#sys.path.append(top)
#cog.msg("top dir     : %s" % top)
#import glob
#cog.outl("recognizers = []")
#for filename in glob.iglob(os.path.join(path, "*.py")):
#    if filename.endswith("__init__.py"):
#        continue
#    modname = filename.split(".py")[0].split("/")[-1]
#    module = imp.load_source(modname, filename)
#    members = inspect.getmembers(module, inspect.isclass)
#    names = []
#    for name, cls in members:
#        if hasattr(cls, 'identify'):
#            # make sure class is from this module and not an imported dependency
#            if cls.__module__.startswith(modname):
#                names.append(name)
#    if names:
#       cog.outl("from %s import %s" % (modname, ", ".join(names)))
#       for name in names:
#           cog.outl("recognizers.append(%s())" % name)
# ]]]*/
recognizers = []
from image import ImageRecognizer
recognizers.append(ImageRecognizer())
from text import PlainTextRecognizer, PoundBangTextRecognizer, XMLTextRecognizer
recognizers.append(PlainTextRecognizer())
recognizers.append(PoundBangTextRecognizer())
recognizers.append(XMLTextRecognizer())
from meta import OmnivoreRecognizer
recognizers.append(OmnivoreRecognizer())
# [[[end]]]

from envisage.api import Plugin
from traits.api import List


class BuiltinFileRecognizerPlugin(Plugin):
    """ A plugin that contributes to the omnivore.file_type.recognizer extension point. """

    #### 'IPlugin' interface ##################################################

    # The plugin's unique identifier.
    id = 'omnivore.file_type.recognizer.builtin'

    # The plugin's name (suitable for displaying to the user).
    name = 'Builtin File Recognizer Plugin'

    # This tells us that the plugin contributes the value of this trait to the
    # 'greetings' extension point.
    recognizer = List(recognizers, contributes_to='omnivore.file_recognizer')


plugins = [BuiltinFileRecognizerPlugin()]



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   04/13/2005
#
#------------------------------------------------------------------------------

""" Defines a drop target editor for the wxPython user interface toolkit. A
    drop target editor handles drag and drop operations as a drop target.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

import wx

# FIXME: ToolkitEditorFactory is a proxy class defined here just for backward
# compatibility. The class has been moved to the
# traitsui.editors.drop_editor file.
from traitsui.editors.drop_editor \
    import ToolkitEditorFactory

from pyface.wx.drag_and_drop \
    import PythonDropTarget, clipboard

from text_editor \
    import SimpleEditor as Editor

from constants \
    import DropColor

#-------------------------------------------------------------------------
#  'SimpleEditor' class:
#-------------------------------------------------------------------------


class SimpleEditor(Editor):
    """ Simple style of drop editor, which displays a read-only text field that
    contains the string representation of the object trait's value.
    """

    # Background color when it is OK to drop objects.
    ok_color = DropColor

    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        if self.factory.readonly:
            self.control = wx.TextCtrl(parent, -1, self.str_value,
                                       style=wx.TE_READONLY)
            self.set_tooltip()
        else:
            super(SimpleEditor, self).init(parent)
        self.control.SetBackgroundColour(self.ok_color)
        self.control.SetDropTarget(PythonDropTarget(self))

    #-------------------------------------------------------------------------
    #  Returns the text representation of a specified object trait value:
    #-------------------------------------------------------------------------

    def string_value(self, value):
        """ Returns the text representation of a specified object trait value.
        """
        if value is None:
            return ''
        return str(value)

    #-------------------------------------------------------------------------
    #  Handles an error that occurs while setting the object's trait value:
    #-------------------------------------------------------------------------

    def error(self, excp):
        """ Handles an error that occurs while setting the object's trait value.
        """
        pass

#----- Drag and drop event handlers: -------------------------------------

    #-------------------------------------------------------------------------
    #  Handles a Python object being dropped on the control:
    #-------------------------------------------------------------------------

    def wx_dropped_on(self, x, y, data, drag_result):
        """ Handles a Python object being dropped on the tree.
        """
        klass = self.factory.klass
        value = data
        if self.factory.binding:
            value = getattr(clipboard, 'node', None)
        if (klass is None) or isinstance(data, klass):
            self._no_update = True
            try:
                if hasattr(value, 'drop_editor_value'):
                    self.value = value.drop_editor_value()
                else:
                    self.value = value
                if hasattr(value, 'drop_editor_update'):
                    value.drop_editor_update(self.control)
                else:
                    self.control.SetValue(self.str_value)
            finally:
                self._no_update = False
            return drag_result

        return wx.DragNone

    #-------------------------------------------------------------------------
    #  Handles a Python object being dragged over the control:
    #-------------------------------------------------------------------------

    def wx_drag_over(self, x, y, data, drag_result):
        """ Handles a Python object being dragged over the tree.
        """
        if self.factory.binding:
            data = getattr(clipboard, 'node', None)
        try:
            self.object.base_trait(self.name).validate(self.object,
                                                       self.name, data)
            return drag_result
        except:
            return wx.DragNone

# Define the Text and ReadonlyEditor for use by the editor factory.
TextEditor = ReadonlyEditor = SimpleEditor

### EOF ##################################################################



""" Preferences for all help examples.

    :Copyright: 2008, Enthought Inc.
    :License: BSD
    :Author: Janet Swisher
"""
# This software is provided without warranty under the terms of the BSD
# license included in AppTools/trunk/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!


from apptools.preferences.api import PreferencesHelper
from traits.api import Either, Enum, File, Str, provides

# This module's package.
PKG = '.'.join(__name__.split('.')[:-1])

class ExamplesPreferences(PreferencesHelper):
    """ Preferences for all help examples.
    """

    #### Preferences ######################################

    # The path to the preferences
    preferences_path = PKG + '.Examples'

    # The external program to use to view the document, if editor_choice is
    # 'external'. It is a command to run, which may be in the program search
    # path of the current environment, or an absolute path to a program.
    editor = Str




#Enthought imports
from traits.api import HasTraits, Int, Str

class FactoryDefinition(HasTraits):
    """
    A project factory definition.

    An instance of the specified class is used to open and/or create new
    projects.

    The extension with the highest priority wins!  In the event of a tie,
    the first instance wins.

    """

    # The name of the class that implements the factory.
    class_name = Str

    # The priority of this factory
    priority = Int



#!/usr/bin/env python
from fs.commands.runner import Command
import sys
from datetime import datetime

class FSInfo(Command):
     
    usage = """fsinfo [OPTION]... [PATH]
Display information regarding an FS resource"""
     
    def get_optparse(self):
        optparse = super(FSInfo, self).get_optparse()
        optparse.add_option('-k', '--key', dest='keys', action='append', default=[],
                            help='display KEYS only')
        optparse.add_option('-s', '--simple', dest='simple', action='store_true', default=False,
                            help='info displayed in simple format (no table)')
        optparse.add_option('-o', '--omit', dest='omit', action='store_true', default=False,
                            help='omit path name from output')
        optparse.add_option('-d', '--dirsonly', dest='dirsonly', action="store_true", default=False,
                            help="list directories only", metavar="DIRSONLY")
        optparse.add_option('-f', '--filesonly', dest='filesonly', action="store_true", default=False,
                            help="list files only", metavar="FILESONLY")
        return optparse   
        
        
    def do_run(self, options, args):
        
        def wrap_value(val):            
            if val.rstrip() == '\0':
                return self.wrap_error('... missing ...')
            return val

        def make_printable(text):
            if not isinstance(text, basestring):
                try:
                    text = str(text)
                except:
                    try:
                        text = unicode(text)
                    except:
                        text = repr(text)
            return text
                
                
        keys = options.keys or None
        for fs, path, is_dir in self.get_resources(args,
                                                   files_only=options.filesonly,
                                                   dirs_only=options.dirsonly):                        
            if not options.omit:
                if options.simple:           
                    file_line = u'%s\n' % self.wrap_filename(path)
                else:
                    file_line = u'[%s] %s\n' % (self.wrap_filename(path), self.wrap_faded(fs.desc(path)))
                self.output(file_line)            
            info = fs.getinfo(path)
                                            
            for k, v in info.items():
                if k.startswith('_'):
                    del info[k]
                elif not isinstance(v, (basestring, int, long, float, bool, datetime)):
                    del info[k]                
                        
            if keys:            
                table = [(k, make_printable(info.get(k, '\0'))) for k in keys]
            else:
                keys = sorted(info.keys())
                table = [(k, make_printable(info[k])) for k in sorted(info.keys())]
                                 
            if options.simple:
                for row in table:
                    self.output(row[-1] + '\n')
            else:                                        
                self.output_table(table, {0:self.wrap_table_header, 1:wrap_value})            

        
def run():
    return FSInfo().run()         
    
if __name__ == "__main__":
    sys.exit(run())    



import threading
import Queue as queue

def make_job(job_callable, *args, **kwargs):
    """ Returns a callable that calls the supplied callable with given arguements. """
    def job():
        return job_callable(*args, **kwargs)
    return job


class _PoolThread(threading.Thread):
    """ Internal thread class that runs jobs. """
    
    def __init__(self, queue, name):
        super(_PoolThread, self).__init__()
        self.queue = queue
        self.name = name
        
    def __str__(self):
        return self.name
        
    def run(self):
                
        while True:
            try:
                _priority, job = self.queue.get()                
            except queue.Empty:                
                break
            
            if job is None:            
                break
            
            if callable(job):
                try:
                    job()                
                except Exception, e:
                    print e                    
            self.queue.task_done()
    

class ThreadPool(object):
    
    def __init__(self, num_threads, size=None, name=''):
        
        self.num_threads = num_threads
        self.name = name
        self.queue = queue.PriorityQueue(size)
        self.job_no = 0
        
        self.threads =  [_PoolThread(self.queue, '%s #%i' % (name, i)) for i in xrange(num_threads)]
        
        for thread in self.threads:
            thread.start()
                
    def _make_priority_key(self, i):
        no = self.job_no
        self.job_no += 1
        return (i, no)

    def job(self, job_callable, *args, **kwargs):
        """ Post a job to the queue. """
        def job():
            return job_callable(*args, **kwargs)
        self.queue.put( (self._make_priority_key(1), job), True )
        return self.job_no        
        
    def flush_quit(self):
        """ Quit after all tasks on the queue have been processed. """        
        for thread in self.threads:
            self.queue.put( (self._make_priority_key(1), None) )        
        for thread in self.threads:
            thread.join()
        
    def quit(self):
        """ Quit as soon as possible, potentially leaving tasks on the queue. """
        for thread in self.threads:
            self.queue.put( (self._make_priority_key(0), None) )            
        for thread in self.threads:
            thread.join()


if __name__ == "__main__":
    import time
    

    def job(n):
        print "Starting #%i" % n
        time.sleep(1)
        print "Ending #%i" % n
    
    pool = ThreadPool(5, 'test thread')
    
    for n in range(20):
        pool.job(job, n)
    
    pool.flush_quit()

        
    


#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought logger package component>
#------------------------------------------------------------------------------

# Standard library imports.
from logging import Handler

# Local imports.
from ring_buffer import RingBuffer


class LogQueueHandler(Handler):

    """ Buffers up the log messages so that we can display them later.
        This is important on startup when log messages are generated before
        the ui has started.  By putting them in this queue we can display
        them once the ui is ready.
    """

    # The view where updates will go
    _view = None


    def __init__(self, size=1000):
        Handler.__init__(self)
        # only buffer 1000 log records
        self.size = size
        self.ring = RingBuffer(self.size)
        self.dirty = False
        return


    def emit(self, record):
        """ Actually this is more like an enqueue than an emit()."""
        self.ring.append(record)
        if self._view is not None:
            try:
                self._view.update()
            except Exception, e:
                pass
        self.dirty = True
        return


    def get(self):
        self.dirty = False

        try:
            result = self.ring.get()
        except Exception, msg:
            # we did our best and it won't cause too much damage
            # to just return a bogus message
            result = []

        return result


    def has_new_records(self):
        return self.dirty


    def reset(self):
        # start over with a new empty buffer
        self.ring = RingBuffer(self.size)
        if self._view is not None:
            try:
                self._view.update()
            except Exception, e:
                pass
        self.dirty = True
        return


## EOF ##################################################################



#-----------------------------------------------------------------------------
#
#  Copyright (c) 2006 by Enthought, Inc.
#  All rights reserved.
#
#  Author: Dave Peterson <dpeterson@enthought.com>
#
#-----------------------------------------------------------------------------

""" Tests the updater functionality of the sweet_pickle framework.
"""

# Standard library imports.
import unittest
import logging

# Enthought library imports
import apptools.sweet_pickle as sweet_pickle
from apptools.sweet_pickle.global_registry import _clear_global_registry


logger = logging.getLogger(__name__)


##############################################################################
# class 'UpdaterTestCase'
##############################################################################

class UpdaterTestCase(unittest.TestCase):
    """ Tests the updater functionality of the sweet_pickle framework.
    """

    ##########################################################################
    # 'TestCase' interface
    ##########################################################################

    ### public interface #####################################################

    def setUp(self):
        """ Creates the test fixture.

            Overridden here to ensure each test starts with an empty global
            registry.
        """
        # Clear the global registry
        _clear_global_registry()

        # Cache a reference to the new global registry
        self.registry = sweet_pickle.get_global_registry()


    ##########################################################################
    # 'UpdaterTestCase' interface
    ##########################################################################

    ### public interface #####################################################

    def test_add_mapping(self):
        """ Validates the behavior of the add_mapping function.
        """
        # Add a single mapping and validate that all it did was add a class
        # mapping and that the mapping has the expected values.
        key = ('foo', 'Foo')
        value = ('bar', 'Bar')
        self.registry.add_mapping(key[0], key[1], value[0], value[1])
        self._validate_exactly_one_class_map(key, value)

        # Overwrite with a new mapping and validate the state is what we
        # expect.
        value = ('baz', 'Baz')
        self.registry.add_mapping(key[0], key[1], value[0], value[1])
        self._validate_exactly_one_class_map(key, value)


    def test_add_mapping_to_class(self):
        """ Validates the behavior of the add_mapping_to_class function.
        """
        # Add a single mapping and validate that all it did was add a class
        # mapping and that the mapping has the expected values.
        key = ('foo', 'Foo')
        class Bar:
            pass
        value = (Bar.__module__, 'Bar')
        self.registry.add_mapping_to_class(key[0], key[1], Bar)
        self._validate_exactly_one_class_map(key, value)

        # Overwrite with a new mapping and validate the state is what we
        # expect.
        class Baz:
            pass
        value = (Baz.__module__, 'Baz')
        self.registry.add_mapping_to_class(key[0], key[1], Baz)
        self._validate_exactly_one_class_map(key, value)


    def test_add_mappings(self):
        """ Validates the behavior of the add_mappings function.
        """
        # Add a single mapping and validate that all it did was add a class
        # mapping and that the mapping has the expected values
        key = ('foo', 'Foo')
        value = ('bar', key[1])
        names = [key[1]]
        self.registry.add_mappings(key[0], value[0], names)
        self._validate_exactly_one_class_map(key, value)

        # Add multiple mappings and validate that the registry has the expected
        # values.
        key = ('foo', 'Foo')
        value = ('bar', 'Bar')
        names = ['Foo', 'Bar', 'Baz', 'Enthought']
        self.registry.add_mappings(key[0], value[0], names)
        self.assertEqual(len(names), len(self.registry.class_map))
        self.assertEqual(0, len(self.registry.state_functions))
        self.assertEqual(0, len(self.registry.version_attribute_map))
        self.assertEqual(0, len(self.registry._state_function_classes))
        items = []
        for n in names:
            items.append( ((key[0], n), (value[0], n)) )
        self._validate_class_map_contents(items)


    def test_add_state_function(self):
        """ Validates the behavior of the add_state_function function.
        """
        # Add a single function and validate that all it did was add a state
        # mapping and that the mapping has the expected values.
        key = ('foo', 'Foo', 1)
        def fn():
            pass
        self.registry.add_state_function(key[0], key[1], key[2], fn)
        self._validate_exactly_one_state_function(key, [fn])

        # Add an additional function for the same state and validate the state
        # is what we expect.
        def fn2():
            pass
        self.registry.add_state_function(key[0], key[1], key[2], fn2)
        self._validate_exactly_one_state_function(key, [fn, fn2])

        # Add a state function for another version of the same class and
        # validate that all the values are as expected.
        key2 = ('foo', 'Foo', 2)
        self.registry.add_state_function(key2[0], key2[1], key2[2], fn2)
        self.assertEqual(0, len(self.registry.class_map))
        self.assertEqual(2, len(self.registry.state_functions))
        self.assertEqual(0, len(self.registry.version_attribute_map))
        self.assertEqual(1, len(self.registry._state_function_classes))
        self._validate_state_function_contents(
            [(key, [fn, fn2]), (key2, [fn2])],
            {(key[0], key[1]): 3}
            )


    def test_add_state_function_for_class(self):
        """ Validates the behavior of the add_state_function_for_class function.
        """
        # Add a single function and validate that all it did was add a state
        # mapping and that the mapping has the expected values.
        class Bar:
            pass
        key = (Bar.__module__, 'Bar', 1)
        def fn():
            pass
        self.registry.add_state_function_for_class(Bar, key[2], fn)
        self._validate_exactly_one_state_function(key, [fn])

        # Add an additional function for the same state and validate the state
        # is what we expect.
        def fn2():
            pass
        self.registry.add_state_function_for_class(Bar, key[2], fn2)
        self._validate_exactly_one_state_function(key, [fn, fn2])

        # Add a state function for another class and validate that all the
        # values are as expected.
        class Baz:
            pass
        key2 = (Baz.__module__, 'Baz', 2)
        self.registry.add_state_function_for_class(Baz, key2[2], fn2)
        self.assertEqual(0, len(self.registry.class_map))
        self.assertEqual(2, len(self.registry.state_functions))
        self.assertEqual(0, len(self.registry.version_attribute_map))
        self.assertEqual(2, len(self.registry._state_function_classes))
        self._validate_state_function_contents(
            [(key, [fn, fn2]), (key2, [fn2])],
            {(key[0], key[1]): 2, (key2[0], key2[1]): 1}
            )


    def test_merge_updater(self):
        """ Validates the behavior of the merge_updater function.
        """
        # Merge in one update and validate the state of the registry is as
        # expected.
        def fn1():
            pass
        updater = sweet_pickle.Updater(
            class_map = {
                ('foo', 'Foo'): ('foo.bar', 'Foo'),
                },
            state_functions = {
                ('foo', 'Foo', 1): [fn1],
                },
            version_attribute_map = {
                ('foo', 'Foo'): 'version',
                },
            )
        self.registry.merge_updater(updater)
        self.assertEqual(1, len(self.registry.class_map))
        self.assertEqual(1, len(self.registry.state_functions))
        self.assertEqual(1, len(self.registry.version_attribute_map))
        self.assertEqual(1, len(self.registry._state_function_classes))
        self._validate_class_map_contents(updater.class_map.items())
        counts = {('foo', 'Foo'): 1}
        self._validate_state_function_contents(updater.state_functions.items(),
            counts)

        # Merge in a second updater and validate the state of the registry is
        # as expected.
        def fn2():
            pass
        updater2 = sweet_pickle.Updater(
            class_map = {
                ('foo.bar', 'Foo'): ('bar', 'Bar'),
                ('bar', 'Bar'): ('foo.bar.baz', 'Baz'),
                },
            state_functions = {
                ('foo', 'Foo', 1): [fn2],
                ('foo', 'Foo', 2): [fn2],
                },
            version_attribute_map = {
                ('foo.bar', 'Foo'): '_version',
                },
            )
        self.registry.merge_updater(updater2)
        self.assertEqual(3, len(self.registry.class_map))
        self.assertEqual(2, len(self.registry.state_functions))
        self.assertEqual(2, len(self.registry.version_attribute_map))
        self.assertEqual(1, len(self.registry._state_function_classes))
        self._validate_class_map_contents(updater.class_map.items() + \
            updater2.class_map.items())
        counts = {('foo', 'Foo'): 3}
        self._validate_state_function_contents(
            [ (('foo', 'Foo', 1), [fn1, fn2]), (('foo', 'Foo', 2), [fn2]) ],
            counts)


    def test_registry_starts_empty(self):
        """ Validates that the registry is starting empty for each test.
        """
        self.assertEqual(0, len(self.registry.class_map))
        self.assertEqual(0, len(self.registry.state_functions))
        self.assertEqual(0, len(self.registry.version_attribute_map))
        self.assertEqual(0, len(self.registry._state_function_classes))


    ### protected interface ##################################################

    def _validate_class_map_contents(self, items):
        """ Validates that the registry's class_map contains the specified
            items.
        """
        for key, value in items:
            self.assertEqual(True, key in self.registry.class_map,
                'Key ' + str(key) + ' not in class_map')
            self.assertEqual(value, self.registry.class_map[key],
                str(value) + ' != ' + str(self.registry.class_map[key]) + \
                ' for key ' + str(key))
            self.assertEqual(True,
                self.registry.has_class_mapping(key[0], key[1]),
                'Registry reports no class mapping for key ' + str(key))


    def _validate_exactly_one_class_map(self, key, value):
        """ Validates that the registry has exactly one class_map entry
            with the specified key and value.
        """
        self.assertEqual(1, len(self.registry.class_map))
        self.assertEqual(0, len(self.registry.state_functions))
        self.assertEqual(0, len(self.registry.version_attribute_map))
        self.assertEqual(0, len(self.registry._state_function_classes))
        self._validate_class_map_contents([(key, value)])


    def _validate_exactly_one_state_function(self, key, value):
        """ Validates that the registry has exactly one state_function entry
            with the specified key and value.
        """
        self.assertEqual(0, len(self.registry.class_map))
        self.assertEqual(1, len(self.registry.state_functions))
        self.assertEqual(0, len(self.registry.version_attribute_map))
        self.assertEqual(1, len(self.registry._state_function_classes))
        self.assertEqual(key, self.registry.state_functions.keys()[0])
        self.assertEqual(value, self.registry.state_functions[key])
        classes_key = (key[0], key[1])
        self.assertEqual(classes_key,
            self.registry._state_function_classes.keys()[0])
        self.assertEqual(len(value),
            self.registry._state_function_classes[classes_key])


    def _validate_state_function_contents(self, items, counts):
        """ Validates that the registry's state functions contains the
            specified items and the class count matches the specified count.
        """
        for key, value in items:
            self.assertEqual(True, key in self.registry.state_functions,
                'Key ' + str(key) + ' not in state functions')
            self.assertEqual(value, self.registry.state_functions[key],
                str(value) + ' != ' + \
                str(self.registry.state_functions[key]) + ' for key ' + \
                str(key))
            self.assertEqual(True, self.registry.has_state_function(
                key[0], key[1]),
                'Registry reports no state function for key ' + str(key))

            classes_key = (key[0], key[1])
            count = counts[classes_key]
            self.assertEqual(count,
                self.registry._state_function_classes[classes_key])


if __name__ == "__main__":
    unittest.main()


### EOF ######################################################################



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   11/06/2004
#
#------------------------------------------------------------------------------

""" Adds a "category" capability to Traits-based classes, similar to that
    provided by the Cocoa (Objective-C) environment for the Macintosh.

    You can use categories to extend an existing HasTraits class, as an
    alternative to subclassing. An advantage of categories over subclassing is
    that you can access the added members on instances of the original class,
    without having to change them to instances of a subclass. Unlike
    subclassing, categories do not allow overriding trait attributes.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from __future__ import absolute_import

from .has_traits import MetaHasTraits, MetaHasTraitsObject

#-------------------------------------------------------------------------------
#  'MetaCategory' class:
#-------------------------------------------------------------------------------

class MetaCategory ( MetaHasTraits ):

    def __new__ ( cls, class_name, bases, class_dict ):

        # Make sure the correct usage is being applied:
        if len( bases ) > 2:
            raise TypeError, \
                  "Correct usage is: class FooCategory(Category,Foo):"

        # Process any traits-related information in the class dictionary:
        MetaCategoryObject( cls, class_name, bases, class_dict, True )

        # Move all remaining items in our class dictionary to the base class's
        # dictionary:
        if len( bases ) == 2:
            category_class = bases[1]
            for name, value in class_dict.items():
                if not hasattr( category_class, name ):
                    setattr( category_class, name, value )
                    del class_dict[ name ]

        # Finish building the class using the updated class dictionary:
        return type.__new__( cls, class_name, bases, class_dict )

#-------------------------------------------------------------------------------
#  'MetaCategoryObject' class:
#-------------------------------------------------------------------------------

class MetaCategoryObject ( MetaHasTraitsObject ):

    #---------------------------------------------------------------------------
    #  Adds the traits meta-data to the class:
    #---------------------------------------------------------------------------

    def add_traits_meta_data ( self, bases, class_dict, base_traits,
                               class_traits, instance_traits, prefix_traits,
                               listeners, view_elements ):
        if len( bases ) == 2:
            # Update the class and each of the existing subclasses:
            bases[1]._add_trait_category( base_traits, class_traits,
                      instance_traits, prefix_traits, listeners,
                      view_elements )
        else:
            MetaHasTraitsObject.add_traits_meta_data( self, bases,
                   class_dict, base_traits, class_traits, instance_traits,
                   prefix_traits, listeners, view_elements )

#-------------------------------------------------------------------------------
#  'Category' class:
#-------------------------------------------------------------------------------

class Category ( object ):
    """ Used for defining "category" extensions to existing classes.

    To define a class as a category, specify "Category," followed by the name
    of the base class name in the base class list.

    The following example demonstrates defining a category::

        from traits.api import HasTraits, Str, Category

        class Base(HasTraits):
            x = Str("Base x")
            y = Str("Base y")

        class BaseExtra(Category, Base):
            z = Str("BaseExtra z")
    """

    __metaclass__ = MetaCategory




""" A simple traits-aware code browser. """


# Standard library imports.
import imp, logging, os, stat, warnings

# Enthought library imports.
from apptools.io.api import File
from traits.api import Any, Bool, Event, HasTraits, Str

# Local imports.
from .module import ModuleFactory
from .package import Package
from ..._compat import pickle

# Logging.
logger = logging.getLogger(__name__)


# Filter future warnings for maxint, since it causes warnings when compiling
# anything that has RGBA colors defined as hex values.
warnings.filterwarnings(
    "ignore", r'hex/oct constants > sys\.maxint .*', FutureWarning
)


class CodeBrowser(HasTraits):
    """ A simple, traits-aware code browser. """

    #### 'ClassBrowser' interface #############################################

    # The filename of the (optional, persistent) code 'database'.
    filename = Str

    #### Events ####

    # Fired when a module is about to be parsed.
    parsing_module = Event

    # Fired when a module has been parsed.
    parsed_module = Event

    #### Private interface ####################################################

    # The code 'database' that contains every module that has been parsed.
    _database = Any

    # Has the code database been changed (i.e., do we need to save it)?
    _database_changed = Bool(False)

    ###########################################################################
    # 'CodeBrowser' interface.
    ###########################################################################

    def load(self):
        """ Load the code browser 'database'. """

        # If a persisted code database exists then load it...
        if os.path.isfile(self.filename):
            logger.debug('loading code database...')

            f = open(self.filename, 'rb')
            self._database = pickle.load(f)
            f.close()

            logger.debug('code database loaded.')

        # ... otherwise we have a nice, fresh and very empty database.
        else:
            self._database = {}

        return

    def save(self):
        """ Save the code browser 'database' to disk. """

        if self._database_changed:
            logger.debug('saving code database...')

            f = open(self.filename, 'wb')
            pickle.dump(self._database, f, 1)
            f.close()

            self._database_changed = False

            logger.debug('code database saved.')

        else:
            logger.debug('code database unchanged - nothing saved.')

        return

    def read_package(self, package_name):
        """ Parse every module in the specified package. """

        filename = self.find_module(package_name)
        if filename is None:
            raise ValueError("no such package '%s'" % package_name)

        package = Package(filename=filename, name=package_name)
        self.read_directory(filename, package)

        return package

    def read_directory(self, filename, package=None):
        """ Parse every module in the specified directory. """

        directory = File(filename)
        if not directory.is_folder:
            raise ValueError("%s is NOT a directory." % filename)

        if package is not None:
            contents = package.contents

        else:
            contents = []

        for child in directory.children:
            # If the child is a Python file then parse it.
            if child.ext == '.py':
                contents.append(self.read_file(child.path, package))

            # If the child is a sub-package then recurse!
            elif child.is_package:
                if package is not None:
                    sub_package_name = '%s.%s' % (package.name, child.name)
                    sub_package = Package(
                        filename = child.path,
                        name     = sub_package_name,
                        parent   = package
                    )

                else:
                    sub_package = Package(filename=child.path, name=child.name)

                self.read_directory(child.path, sub_package)
                contents.append(sub_package)

        return contents

    def read_file(self, filename, namespace=None):
        """ Parse a file. """

        # Only parse the file if we haven't parsed it before or it has been
        # modified since we last parsed it!
        module, mod_time = self._database.get(filename, (None, None))
        if module is None or mod_time != os.stat(filename)[stat.ST_MTIME]:
            # Event notification.
            self.parsing_module = filename
            logger.debug('parsing module %s' % filename)

            module_factory = ModuleFactory()
            try:
                module = module_factory.from_file(filename, namespace)

                # Event notification.
                self.parsed_module = module
                logger.debug('parsed module %s' % filename)

                # Add the parsed module to the database.
                self._database[filename] = (
                    module, os.stat(filename)[stat.ST_MTIME]
                )
                self._database_changed = True

            except SyntaxError:
                logger.debug('error parsing module %s' % filename)

        return module

##     def read_module(self, module_name):
##         """ Parses a module. """

##         filename = self.find_module(module_name)
##         if filename is not None:
##             module = self.read_file(filename)

##         else:
##             module = None

##         return module

    def find_module(self, module_name, path=None):
        """ Return the filename for the specified module. """

        components = module_name.split('.')

        try:
            # Look up the first component of the module name (of course it
            # could be the *only* component).
            f, filename, description = imp.find_module(components[0], path)

            # If the module is in a package then go down each level in the
            # package hierarchy in turn.
            if len(components) > 0:
                for component in components[1:]:
                    f, filename, description = imp.find_module(
                        component, [filename]
                    )

        except ImportError:
            filename = None

        return filename

    ###########################################################################
    # Private interface.
    ###########################################################################

    #### Trait initializers ###################################################

    def __database_default(self):
        """ Trait initializer. """

        return {}

    #### Trait change handlers ################################################

    def _filename_changed(self):
        """ Called when the filename of the code database is changed. """

        # Load the contents of the database.
        self.load()

        return

#### EOF ######################################################################



import platform
import logging

import fs.errors as errors
from fs import SEEK_END

python3 = int(platform.python_version_tuple()[0]) > 2

if python3:
    from urllib.parse import urlencode, pathname2url, quote
    from urllib.request import Request, urlopen
else:
    from urllib import urlencode, pathname2url
    from urllib2 import Request, urlopen, quote

class PutRequest(Request):
    def __init__(self, *args, **kwargs):
        self.get_method = lambda: u'PUT'
        Request.__init__(self, *args, **kwargs)
         
class DeleteRequest(Request):
    def __init__(self, *args, **kwargs):
        self.get_method = lambda: u'DELETE'
        Request.__init__(self, *args, **kwargs)
              
class Connection:
    def __init__(self, webapi):
        self.webapi = webapi
        self.headers = {'Accept': 'text/plain'}

    def _get_headers(self, f, size=None):
        '''
            Retrieve length of string or file object and prepare HTTP headers.
        '''
        if isinstance(f, basestring):
            # Just set up content length
            size = len(f)
        elif getattr(f, 'read', None):
            if size == None:
                # When size is already known, skip this           
                f.seek(0, SEEK_END)
                size = f.tell()
                f.seek(0)
        else:
            raise errors.UnsupportedError("Cannot handle type %s" % type(f))
        
        headers = {'Content-Length': size}
        headers.update(self.headers)
        return headers

    def _urlencode(self, data):
        _data = {}
        for k, v in data.items():
            _data[k.encode('utf-8')] = v.encode('utf-8')
        return urlencode(_data)

    def _quotepath(self, path, params={}):
        q = quote(path.encode('utf-8'), safe='/')
        if params:
            return u"%s?%s" % (q, self._urlencode(params))
        return q
        
    def _urlopen(self, req):
        try:
            return urlopen(req)
        except Exception, e:
            if not getattr(e, 'getcode', None):
                raise errors.RemoteConnectionError(str(e))
            code = e.getcode()
            if code == 500:
                # Probably out of space or unhappiness error
                raise errors.StorageSpaceError(e.fp.read())
            elif code in (400, 404, 410):
                # Standard not found
                raise errors.ResourceNotFoundError(e.fp.read())
            raise errors.ResourceInvalidError(e.fp.read())
        
    def post(self, path, data={}, params={}):
        data = self._urlencode(data)
        path = self._quotepath(path, params)
        req = Request(''.join([self.webapi, path]), data, headers=self.headers)
        return self._urlopen(req)
    
    def get(self, path, data={}, offset=None, length=None):
        data = self._urlencode(data)
        path = self._quotepath(path)
        if data: 
            path = u'?'.join([path, data])

        headers = {}
        headers.update(self.headers)
        if offset:
            if length:
                headers['Range'] = 'bytes=%d-%d' % \
                                    (int(offset), int(offset+length))
            else:
                headers['Range'] = 'bytes=%d-' % int(offset)
            
        req = Request(''.join([self.webapi, path]), headers=headers)
        return self._urlopen(req)

    def put(self, path, data, size=None, params={}):
        path = self._quotepath(path, params)
        headers = self._get_headers(data, size=size)
        req = PutRequest(''.join([self.webapi, path]), data, headers=headers)    
        return self._urlopen(req)
    
    def delete(self, path, data={}):   
        path = self._quotepath(path)
        req = DeleteRequest(''.join([self.webapi, path]), data, headers=self.headers)
        return self._urlopen(req)



#------------------------------------------------------------------------------
# Copyright (c) 2006, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------

# local imports
from .checkbox_image_renderer import CheckboxImageRenderer
from .grid_cell_renderer import GridCellRenderer

class CheckboxRenderer(GridCellRenderer):

    def __init__(self, **traits):

        # base-class constructor
        super(CheckboxRenderer, self).__init__(**traits)

        # initialize the renderer, if it hasn't already been initialized
        if self.renderer is None:
            self.renderer = CheckboxImageRenderer()

        return

    def on_left_click(self, grid, row, col):
        """ Toggles the value. """

        value = grid.model.get_cell_drag_value(row, col)

        try:
            grid.model.set_value(row, col, not value)
        except:
            pass

        return True

#### EOF ######################################################################



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#
#------------------------------------------------------------------------------

""" Table column object for Color traits.
"""

import warnings
warnings.warn("traitsui.wx.color_column is deprecated. Use the "
              "backend-independent implementation in traitsui.color_column")

from traitsui.color_column import ColorColumn



""" The default resource manager. """


# Enthought library imports.
from traits.api import Dict, HasTraits, Str, provides

# Local imports.
from .i_resource_manager import IResourceManager
from .i_resource_protocol import IResourceProtocol


@provides(IResourceManager)
class ResourceManager(HasTraits):
    """ The default resource manager. """

    #### 'IResourceManager' interface #########################################

    # The protocols used by the manager to resolve resource URLs.
    resource_protocols = Dict(Str, IResourceProtocol)

    ###########################################################################
    # 'IResourceManager' interface.
    ###########################################################################

    #### Trait initializers ###################################################

    def _resource_protocols_default(self):
        """ Trait initializer. """

        # We do the import(s) here in case somebody wants a resource manager
        # that doesn't use the default protocol(s).
        from .file_resource_protocol import FileResourceProtocol
        from .http_resource_protocol import HTTPResourceProtocol
        from .package_resource_protocol import PackageResourceProtocol

        resource_protocols = {
            'file'    : FileResourceProtocol(),
            'http'    : HTTPResourceProtocol(),
            'pkgfile' : PackageResourceProtocol()
        }

        return resource_protocols

    #### Methods ##############################################################

    def file(self, url):
        """ Return a readable file-like object for the specified url. """

        protocol_name, address = url.split('://')

        protocol = self.resource_protocols.get(protocol_name)
        if protocol is None:
            raise ValueError('unknown protocol in URL %s' % url)

        return protocol.file(address)

#### EOF ######################################################################




from contextlib import contextmanager
import warnings

from traits.testing.unittest_tools import unittest
from traits.trait_base import ETSConfig
import traitsui.toolkit


@contextmanager
def clear_toolkit():
    """ If a toolkit has been selected, clear it, resetting on exit """
    old_ETS_toolkit = ETSConfig._toolkit
    old_traitsui_toolkit = traitsui.toolkit._toolkit
    ETSConfig._toolkit = ''
    traitsui.toolkit._toolkit = None
    try:
        yield
    finally:
        ETSConfig._toolkit = old_ETS_toolkit
        traitsui.toolkit._toolkit = old_traitsui_toolkit


class TestToolkit(unittest.TestCase):

    def test_default_toolkit(self):
        with clear_toolkit():
            # try to import default toolkit - this is just a smoke test
            tk = traitsui.toolkit.toolkit()

            self.assertNotEqual(ETSConfig.toolkit, '')
            self.assertIsInstance(tk, traitsui.toolkit.Toolkit)

    def test_nonstandard_toolkit(self):
        with clear_toolkit():
            # try to import a non-default toolkit
            tk = traitsui.toolkit.toolkit('null')

            self.assertEqual(ETSConfig.toolkit, 'null')
            from traitsui.null import toolkit
            self.assertIs(tk, toolkit)

    def test_nonexistent_toolkit(self):
        with clear_toolkit():
            # try to import a non-existent toolkit
            tk = traitsui.toolkit.toolkit('nosuchtoolkit')

            # should fail, and give us 'null'
            self.assertEqual(ETSConfig.toolkit, 'null')
            from traitsui.null import toolkit
            self.assertIs(tk, toolkit)



# =============================================================================
# OWSLib. Copyright (C) 2005 Sean C. Gillies
#
# Contact email: sgillies@frii.com
#
# $Id: wfs.py 503 2006-02-01 17:09:12Z dokai $
# =============================================================================

from __future__ import (absolute_import, division, print_function)

#owslib imports:
from owslib.ows import ServiceIdentification, ServiceProvider, OperationsMetadata
from owslib.etree import etree
from owslib.util import nspath, testXMLValue, openURL
from owslib.crs import Crs
from owslib.feature import WebFeatureService_
from owslib.namespaces import Namespaces

#other imports
import cgi
from six import PY2
from six.moves import cStringIO as StringIO
try:
    from urllib import urlencode
except ImportError:
    from urllib.parse import urlencode

import logging
from owslib.util import log

n = Namespaces()
WFS_NAMESPACE = n.get_namespace("wfs20")
OWS_NAMESPACE = n.get_namespace("ows110")
OGC_NAMESPACE = n.get_namespace("ogc")
GML_NAMESPACE = n.get_namespace("gml")
FES_NAMESPACE = n.get_namespace("fes")


class ServiceException(Exception):
    pass


class WebFeatureService_2_0_0(WebFeatureService_):
    """Abstraction for OGC Web Feature Service (WFS).

    Implements IWebFeatureService.
    """
    def __new__(self,url, version, xml, parse_remote_metadata=False, timeout=30):
        """ overridden __new__ method 
        
        @type url: string
        @param url: url of WFS capabilities document
        @type xml: string
        @param xml: elementtree object
        @type parse_remote_metadata: boolean
        @param parse_remote_metadata: whether to fully process MetadataURL elements
        @param timeout: time (in seconds) after which requests should timeout
        @return: initialized WebFeatureService_2_0_0 object
        """
        obj=object.__new__(self)
        obj.__init__(url, version, xml, parse_remote_metadata, timeout)
        return obj
    
    def __getitem__(self,name):
        ''' check contents dictionary to allow dict like access to service layers'''
        if name in self.__getattribute__('contents').keys():
            return self.__getattribute__('contents')[name]
        else:
            raise KeyError("No content named %s" % name)
    
    
    def __init__(self, url,  version, xml=None, parse_remote_metadata=False, timeout=30):
        """Initialize."""
        if log.isEnabledFor(logging.DEBUG):
            log.debug('building WFS %s'%url)
        self.url = url
        self.version = version
        self.timeout = timeout
        self._capabilities = None
        reader = WFSCapabilitiesReader(self.version)
        if xml:
            self._capabilities = reader.readString(xml)
        else:
            self._capabilities = reader.read(self.url)
        self._buildMetadata(parse_remote_metadata)
    
    def _buildMetadata(self, parse_remote_metadata=False):
        '''set up capabilities metadata objects: '''
        
        #serviceIdentification metadata
        serviceidentelem=self._capabilities.find(nspath('ServiceIdentification'))
        self.identification=ServiceIdentification(serviceidentelem)  
        #need to add to keywords list from featuretypelist information:
        featuretypelistelem=self._capabilities.find(nspath('FeatureTypeList', ns=WFS_NAMESPACE))
        featuretypeelems=featuretypelistelem.findall(nspath('FeatureType', ns=WFS_NAMESPACE))
        for f in featuretypeelems:  
            kwds=f.findall(nspath('Keywords/Keyword',ns=OWS_NAMESPACE))
            if kwds is not None:
                for kwd in kwds[:]:
                    if kwd.text not in self.identification.keywords:
                        self.identification.keywords.append(kwd.text)
	
   
        #TODO: update serviceProvider metadata, miss it out for now
        serviceproviderelem=self._capabilities.find(nspath('ServiceProvider'))
        self.provider=ServiceProvider(serviceproviderelem)   
        
        #serviceOperations metadata 
        self.operations=[]
        
        for elem in self._capabilities.find(nspath('OperationsMetadata'))[:]:
            if elem.tag !=nspath('ExtendedCapabilities'):
                self.operations.append(OperationsMetadata(elem))
                   
        #serviceContents metadata: our assumption is that services use a top-level 
        #layer as a metadata organizer, nothing more. 
        
        self.contents={} 
        featuretypelist=self._capabilities.find(nspath('FeatureTypeList',ns=WFS_NAMESPACE))
        features = self._capabilities.findall(nspath('FeatureTypeList/FeatureType', ns=WFS_NAMESPACE))
        for feature in features:
            cm=ContentMetadata(feature, featuretypelist, parse_remote_metadata)
            self.contents[cm.id]=cm       
        
        #exceptions
        self.exceptions = [f.text for f \
                in self._capabilities.findall('Capability/Exception/Format')]
      
    def getcapabilities(self):
        """Request and return capabilities document from the WFS as a 
        file-like object.
        NOTE: this is effectively redundant now"""
        reader = WFSCapabilitiesReader(self.version)
        return openURL(reader.capabilities_url(self.url), timeout=self.timeout)
    
    def items(self):
        '''supports dict-like items() access'''
        items=[]
        for item in self.contents:
            items.append((item,self.contents[item]))
        return items

    def _makeStringIO(self, strval):
        """
        Helper method to make sure the StringIO being returned will work.

        Differences between Python 2.6/2.7/3.x mean we have a lot of cases to handle.
        """
        if PY2:
            return StringIO(strval)

        return StringIO(strval.decode())

    def getfeature(self, typename=None, filter=None, bbox=None, featureid=None,
                   featureversion=None, propertyname=None, maxfeatures=None,storedQueryID=None, storedQueryParams=None,
                   method='Get', outputFormat=None, startindex=None):
        """Request and return feature data as a file-like object.
        #TODO: NOTE: have changed property name from ['*'] to None - check the use of this in WFS 2.0
        Parameters
        ----------
        typename : list
            List of typenames (string)
        filter : string 
            XML-encoded OGC filter expression.
        bbox : tuple
            (left, bottom, right, top) in the feature type's coordinates == (minx, miny, maxx, maxy)
        featureid : list
            List of unique feature ids (string)
        featureversion : string
            Default is most recent feature version.
        propertyname : list
            List of feature property names. '*' matches all.
        maxfeatures : int
            Maximum number of features to be returned.
        method : string
            Qualified name of the HTTP DCP method to use.
        outputFormat: string (optional)
            Requested response format of the request.
        startindex: int (optional)
            Start position to return feature set (paging in combination with maxfeatures)

        There are 3 different modes of use

        1) typename and bbox (simple spatial query)
        2) typename and filter (==query) (more expressive)
        3) featureid (direct access to known features)
        """
        storedQueryParams = storedQueryParams or {}
        url = data = None
        if typename and type(typename) == type(""):
            typename = [typename]
        if method.upper() == "GET":
            (url) = self.getGETGetFeatureRequest(typename, filter, bbox, featureid,
                                                 featureversion, propertyname,
                                                 maxfeatures, storedQueryID,
                                                 storedQueryParams, outputFormat, 'Get', startindex)
            if log.isEnabledFor(logging.DEBUG):
                log.debug('GetFeature WFS GET url %s'% url)
        else:
            (url,data) = self.getPOSTGetFeatureRequest()


        # If method is 'Post', data will be None here
        u = openURL(url, data, method, timeout=self.timeout)

        # check for service exceptions, rewrap, and return
        # We're going to assume that anything with a content-length > 32k
        # is data. We'll check anything smaller.
        if 'Content-Length' in u.info():
            length = int(u.info()['Content-Length'])
            have_read = False
        else:
            data = u.read()
            have_read = True
            length = len(data)

        if length < 32000:
            if not have_read:
                data = u.read()

            try:
                tree = etree.fromstring(data)
            except BaseException:
                # Not XML
                return self._makeStringIO(data)
            else:
                if tree.tag == "{%s}ServiceExceptionReport" % OGC_NAMESPACE:
                    se = tree.find(nspath('ServiceException', OGC_NAMESPACE))
                    raise ServiceException(str(se.text).strip())
                else:
                    return self._makeStringIO(data)
        else:
            if have_read:
                return self._makeStringIO(data)
            return u


    def getpropertyvalue(self, query=None, storedquery_id=None, valuereference=None, typename=None, method=nspath('Get'),**kwargs):
        ''' the WFS GetPropertyValue method'''
        try:
            base_url = next((m.get('url') for m in self.getOperationByName('GetPropertyValue').methods if m.get('type').lower() == method.lower()))
        except StopIteration:
            base_url = self.url
        request = {'service': 'WFS', 'version': self.version, 'request': 'GetPropertyValue'}
        if query:
            request['query'] = str(query)
        if valuereference: 
            request['valueReference'] = str(valuereference)
        if storedquery_id: 
            request['storedQuery_id'] = str(storedquery_id)
        if typename:
            request['typename']=str(typename)
        if kwargs:
            for kw in kwargs.keys():
                request[kw]=str(kwargs[kw])
        encoded_request=urlencode(request)
        u = openURL(base_url + encoded_request)
        return u.read()
        
        
    def _getStoredQueries(self):
        ''' gets descriptions of the stored queries available on the server '''
        sqs=[]
        #This method makes two calls to the WFS - one ListStoredQueries, and one DescribeStoredQueries. The information is then
        #aggregated in 'StoredQuery' objects
        method=nspath('Get')
        
        #first make the ListStoredQueries response and save the results in a dictionary if form {storedqueryid:(title, returnfeaturetype)}
        try:
            base_url = next((m.get('url') for m in self.getOperationByName('ListStoredQueries').methods if m.get('type').lower() == method.lower()))
        except StopIteration:
            base_url = self.url

        request = {'service': 'WFS', 'version': self.version, 'request': 'ListStoredQueries'}
        encoded_request = urlencode(request)
        u = openURL(base_url, data=encoded_request, timeout=self.timeout)
        tree=etree.fromstring(u.read())
        tempdict={}       
        for sqelem in tree[:]:
            title=rft=id=None
            id=sqelem.get('id')
            for elem in sqelem[:]:
                if elem.tag==nspath('Title', WFS_NAMESPACE):
                    title=elem.text
                elif elem.tag==nspath('ReturnFeatureType', WFS_NAMESPACE):
                    rft=elem.text
            tempdict[id]=(title,rft)        #store in temporary dictionary
        
        #then make the DescribeStoredQueries request and get the rest of the information about the stored queries 
        try:
            base_url = next((m.get('url') for m in self.getOperationByName('DescribeStoredQueries').methods if m.get('type').lower() == method.lower()))
        except StopIteration:
            base_url = self.url
        request = {'service': 'WFS', 'version': self.version, 'request': 'DescribeStoredQueries'}
        encoded_request = urlencode(request)
        u = openURL(base_url, data=encoded_request, timeout=to)
        tree=etree.fromstring(u.read())
        tempdict2={} 
        for sqelem in tree[:]:
            params=[] #list to store parameters for the stored query description
            id =sqelem.get('id')
            for elem in sqelem[:]:
                if elem.tag==nspath('Abstract', WFS_NAMESPACE):
                    abstract=elem.text
                elif elem.tag==nspath('Parameter', WFS_NAMESPACE):
                    newparam=Parameter(elem.get('name'), elem.get('type'))
                    params.append(newparam)
            tempdict2[id]=(abstract, params) #store in another temporary dictionary
        
        #now group the results into StoredQuery objects:
        for key in tempdict.keys(): 
            abstract='blah'
            parameters=[]
            sqs.append(StoredQuery(key, tempdict[key][0], tempdict[key][1], tempdict2[key][0], tempdict2[key][1]))
        return sqs
    storedqueries = property(_getStoredQueries, None)

    def getOperationByName(self, name):
        """Return a named content item."""
        for item in self.operations:
            if item.name == name:
                return item
        raise KeyError("No operation named %s" % name)

class StoredQuery(object):
    '''' Class to describe a storedquery '''
    def __init__(self, id, title, returntype, abstract, parameters):
        self.id=id
        self.title=title
        self.returnfeaturetype=returntype
        self.abstract=abstract
        self.parameters=parameters
        
class Parameter(object):
    def __init__(self, name, type):
        self.name=name
        self.type=type
        
    
class ContentMetadata:
    """Abstraction for WFS metadata.
    
    Implements IMetadata.
    """

    def __init__(self, elem, parent, parse_remote_metadata=False, timeout=30):
        """."""
        self.id = elem.find(nspath('Name',ns=WFS_NAMESPACE)).text
        self.title = elem.find(nspath('Title',ns=WFS_NAMESPACE)).text
        abstract = elem.find(nspath('Abstract',ns=WFS_NAMESPACE))
        if abstract is not None:
            self.abstract = abstract.text
        else:
            self.abstract = None
        self.keywords = [f.text for f in elem.findall(nspath('Keywords',ns=WFS_NAMESPACE))]

        # bboxes
        self.boundingBoxWGS84 = None
        b = elem.find(nspath('WGS84BoundingBox',ns=OWS_NAMESPACE))
        if b is not None:
            lc = b.find(nspath("LowerCorner",ns=OWS_NAMESPACE))
            uc = b.find(nspath("UpperCorner",ns=OWS_NAMESPACE))
            ll = [float(s) for s in lc.text.split()]
            ur = [float(s) for s in uc.text.split()]
            self.boundingBoxWGS84 = (ll[0],ll[1],ur[0],ur[1])

        # there is no such think as bounding box
        # make copy of the WGS84BoundingBox
        self.boundingBox = (self.boundingBoxWGS84[0],
                            self.boundingBoxWGS84[1],
                            self.boundingBoxWGS84[2],
                            self.boundingBoxWGS84[3],
                            Crs("epsg:4326"))
        # crs options
        self.crsOptions = [Crs(srs.text) for srs in elem.findall(nspath('OtherCRS',ns=WFS_NAMESPACE))]
        defaultCrs =  elem.findall(nspath('DefaultCRS',ns=WFS_NAMESPACE))
        if len(defaultCrs) > 0:
            self.crsOptions.insert(0,Crs(defaultCrs[0].text))


        # verbs
        self.verbOptions = [op.tag for op \
            in parent.findall(nspath('Operations/*',ns=WFS_NAMESPACE))]
        self.verbOptions + [op.tag for op \
            in elem.findall(nspath('Operations/*',ns=WFS_NAMESPACE)) \
            if op.tag not in self.verbOptions]
        
        #others not used but needed for iContentMetadata harmonisation
        self.styles=None
        self.timepositions=None
        self.defaulttimeposition=None

        # MetadataURLs
        self.metadataUrls = []
        for m in elem.findall('MetadataURL'):
            metadataUrl = {
                'type': testXMLValue(m.attrib['type'], attrib=True),
                'format': m.find('Format').text.strip(),
                'url': testXMLValue(m.find('OnlineResource').attrib['{http://www.w3.org/1999/xlink}href'], attrib=True)
            }

            if metadataUrl['url'] is not None and parse_remote_metadata:  # download URL
                try:
                    content = openURL(metadataUrl['url'], timeout=timeout)
                    doc = etree.parse(content)
                    try:  # FGDC
                        metadataUrl['metadata'] = Metadata(doc)
                    except:  # ISO
                        metadataUrl['metadata'] = MD_Metadata(doc)
                except Exception:
                    metadataUrl['metadata'] = None

            self.metadataUrls.append(metadataUrl)


class WFSCapabilitiesReader(object):
    """Read and parse capabilities document into a lxml.etree infoset
    """

    def __init__(self, version='2.0.0'):
        """Initialize"""
        self.version = version
        self._infoset = None

    def capabilities_url(self, service_url):
        """Return a capabilities url
        """
        qs = []
        if service_url.find('?') != -1:
            qs = cgi.parse_qsl(service_url.split('?')[1])

        params = [x[0] for x in qs]

        if 'service' not in params:
            qs.append(('service', 'WFS'))
        if 'request' not in params:
            qs.append(('request', 'GetCapabilities'))
        if 'version' not in params:
            qs.append(('version', self.version))

        urlqs = urlencode(tuple(qs))
        return service_url.split('?')[0] + '?' + urlqs

    def read(self, url, timeout=30):
        """Get and parse a WFS capabilities document, returning an
        instance of WFSCapabilitiesInfoset

        Parameters
        ----------
        url : string
            The URL to the WFS capabilities document.
        timeout : number
            A timeout value (in seconds) for the request.
        """
        request = self.capabilities_url(url)
        u = openURL(request, timeout=timeout)
        return etree.fromstring(u.read())

    def readString(self, st):
        """Parse a WFS capabilities document, returning an
        instance of WFSCapabilitiesInfoset

        string should be an XML capabilities document
        """
        if not isinstance(st, str) and not isinstance(st, bytes):
            raise ValueError("String must be of type string or bytes, not %s" % type(st))
        return etree.fromstring(st)



"""
fs.sftpfs
=========

**Currently only avaiable on Python2 due to paramiko not being available for Python3**

Filesystem accessing an SFTP server (via paramiko)

"""

import datetime
import stat as statinfo
import threading
import os
import paramiko
from getpass import getuser
import errno

from fs.base import *
from fs.path import *
from fs.errors import *
from fs.utils import isdir, isfile
from fs import iotools


ENOENT = errno.ENOENT


class WrongHostKeyError(RemoteConnectionError):
    pass


# SFTPClient appears to not be thread-safe, so we use an instance per thread
if hasattr(threading, "local"):
    thread_local = threading.local
    #class TL(object):
    #    pass
    #thread_local = TL
else:
    class thread_local(object):
        def __init__(self):
            self._map = {}

        def __getattr__(self, attr):
            try:
                return self._map[(threading.currentThread().ident, attr)]
            except KeyError:
                raise AttributeError(attr)

        def __setattr__(self, attr, value):
            self._map[(threading.currentThread().ident, attr)] = value


if not hasattr(paramiko.SFTPFile, "__enter__"):
    paramiko.SFTPFile.__enter__ = lambda self: self
    paramiko.SFTPFile.__exit__ = lambda self,et,ev,tb: self.close() and False


class SFTPFS(FS):
    """A filesystem stored on a remote SFTP server.

    This is basically a compatibility wrapper for the excellent SFTPClient
    class in the paramiko module.

    """

    _meta = { 'thread_safe' : True,
              'virtual': False,
              'read_only' : False,
              'unicode_paths' : True,
              'case_insensitive_paths' : False,
              'network' : True,
              'atomic.move' : True,
              'atomic.copy' : True,
              'atomic.makedir' : True,
              'atomic.rename' : True,
              'atomic.setcontents' : False
              }

    def __init__(self,
                 connection,
                 root_path="/",
                 encoding=None,
                 hostkey=None,
                 username='',
                 password=None,
                 pkey=None,
                 agent_auth=True,
                 no_auth=False,
                 look_for_keys=True):
        """SFTPFS constructor.

        The only required argument is 'connection', which must be something
        from which we can construct a paramiko.SFTPClient object.  Possible
        values include:

            * a hostname string
            * a (hostname,port) tuple
            * a paramiko.Transport instance
            * a paramiko.Channel instance in "sftp" mode

        The keyword argument 'root_path' specifies the root directory on the
        remote machine - access to files outside this root will be prevented.

        :param connection: a connection string
        :param root_path: The root path to open
        :param encoding: String encoding of paths (defaults to UTF-8)
        :param hostkey: the host key expected from the server or None if you don't require server validation
        :param username: Name of SFTP user
        :param password: Password for SFTP user
        :param pkey: Public key
        :param agent_auth: attempt to authorize with the user's public keys
        :param no_auth: attempt to log in without any kind of authorization
        :param look_for_keys: Look for keys in the same locations as ssh,
            if other authentication is not succesful

        """
        credentials = dict(username=username,
                           password=password,
                           pkey=pkey)
        self.credentials = credentials

        if encoding is None:
            encoding = "utf8"
        self.encoding = encoding
        self.closed = False
        self._owns_transport = False
        self._credentials = credentials
        self._tlocal = thread_local()
        self._transport = None
        self._client = None

        self.hostname = None
        if isinstance(connection, basestring):
            self.hostname = connection
        elif isinstance(connection, tuple):
            self.hostname = '%s:%s' % connection

        super(SFTPFS, self).__init__()
        self.root_path = abspath(normpath(root_path))

        if isinstance(connection,paramiko.Channel):
            self._transport = None
            self._client = paramiko.SFTPClient(connection)
        else:
            if not isinstance(connection,paramiko.Transport):
                connection = paramiko.Transport(connection)
                connection.daemon = True
                self._owns_transport = True

        if hostkey is not None:
            key = self.get_remote_server_key()
            if hostkey != key:
                raise WrongHostKeyError('Host keys do not match')

        connection.start_client()

        if not connection.is_active():
            raise RemoteConnectionError(msg='Unable to connect')

        if no_auth:
            try:
                connection.auth_none('')
            except paramiko.SSHException:
                pass

        elif not connection.is_authenticated():
            if not username:
                username = getuser()
            try:
                if pkey:
                    connection.auth_publickey(username, pkey)

                if not connection.is_authenticated() and password:
                    connection.auth_password(username, password)

                if agent_auth and not connection.is_authenticated():
                    self._agent_auth(connection, username)

                if look_for_keys and not connection.is_authenticated():
                    self._userkeys_auth(connection, username, password)

                if not connection.is_authenticated():
                    try:
                        connection.auth_none(username)
                    except paramiko.BadAuthenticationType, e:
                        self.close()
                        allowed = ', '.join(e.allowed_types)
                        raise RemoteConnectionError(msg='no auth - server requires one of the following: %s' % allowed, details=e)

                if not connection.is_authenticated():
                    self.close()
                    raise RemoteConnectionError(msg='no auth')

            except paramiko.SSHException, e:
                self.close()
                raise RemoteConnectionError(msg='SSH exception (%s)' % str(e), details=e)

        self._transport = connection

    def __unicode__(self):
        return u'<SFTPFS: %s>' % self.desc('/')

    @classmethod
    def _agent_auth(cls, transport, username):
        """
        Attempt to authenticate to the given transport using any of the private
        keys available from an SSH agent.
        """

        agent = paramiko.Agent()
        agent_keys = agent.get_keys()
        if not agent_keys:
            return None
        for key in agent_keys:
            try:
                transport.auth_publickey(username, key)
                return key
            except paramiko.SSHException:
                pass
        return None

    @classmethod
    def _userkeys_auth(cls, transport, username, password):
        """
        Attempt to authenticate to the given transport using any of the private
        keys in the users ~/.ssh and ~/ssh dirs

        Derived from http://www.lag.net/paramiko/docs/paramiko.client-pysrc.html
        """

        keyfiles = []
        rsa_key = os.path.expanduser('~/.ssh/id_rsa')
        dsa_key = os.path.expanduser('~/.ssh/id_dsa')
        if os.path.isfile(rsa_key):
            keyfiles.append((paramiko.rsakey.RSAKey, rsa_key))
        if os.path.isfile(dsa_key):
            keyfiles.append((paramiko.dsskey.DSSKey, dsa_key))
        # look in ~/ssh/ for windows users:
        rsa_key = os.path.expanduser('~/ssh/id_rsa')
        dsa_key = os.path.expanduser('~/ssh/id_dsa')
        if os.path.isfile(rsa_key):
            keyfiles.append((paramiko.rsakey.RSAKey, rsa_key))
        if os.path.isfile(dsa_key):
            keyfiles.append((paramiko.dsskey.DSSKey, dsa_key))

        for pkey_class, filename in keyfiles:
            key = pkey_class.from_private_key_file(filename, password)
            try:
                transport.auth_publickey(username, key)
                return key
            except paramiko.SSHException:
                pass
        return None

    def __del__(self):
        self.close()

    @synchronize
    def __getstate__(self):
        state = super(SFTPFS,self).__getstate__()
        del state["_tlocal"]
        if self._owns_transport:
            state['_transport'] = self._transport.getpeername()
        return state

    def __setstate__(self,state):
        super(SFTPFS, self).__setstate__(state)
        #for (k,v) in state.iteritems():
        #    self.__dict__[k] = v
        #self._lock = threading.RLock()
        self._tlocal = thread_local()
        if self._owns_transport:
            self._transport = paramiko.Transport(self._transport)
            self._transport.connect(**self._credentials)

    @property
    @synchronize
    def client(self):
        if self.closed:
            return None
        client = getattr(self._tlocal, 'client', None)
        if client is None:
            if self._transport is None:
                return self._client
            client = paramiko.SFTPClient.from_transport(self._transport)
            self._tlocal.client = client
        return client
#        try:
#            return self._tlocal.client
#        except AttributeError:
#            #if self._transport is None:
#            #    return self._client
#            client = paramiko.SFTPClient.from_transport(self._transport)
#            self._tlocal.client = client
#            return client

    @synchronize
    def close(self):
        """Close the connection to the remote server."""
        if not self.closed:
            self._tlocal = None
            #if self.client:
            #    self.client.close()
            if self._owns_transport and self._transport and self._transport.is_active:
                self._transport.close()
            self.closed = True

    def _normpath(self, path):
        if not isinstance(path, unicode):
            path = path.decode(self.encoding)
        npath = pathjoin(self.root_path, relpath(normpath(path)))
        if not isprefix(self.root_path, npath):
            raise PathError(path, msg="Path is outside root: %(path)s")
        return npath

    def getpathurl(self, path, allow_none=False):
        path = self._normpath(path)
        if self.hostname is None:
            if allow_none:
                return None
            raise NoPathURLError(path=path)
        username = self.credentials.get('username', '') or ''
        password = self.credentials.get('password', '') or ''
        credentials = ('%s:%s' % (username, password)).rstrip(':')
        if credentials:
            url = 'sftp://%s@%s%s' % (credentials, self.hostname.rstrip('/'), abspath(path))
        else:
            url = 'sftp://%s%s' % (self.hostname.rstrip('/'), abspath(path))
        return url

    @synchronize
    @convert_os_errors
    @iotools.filelike_to_stream
    def open(self, path, mode='r', buffering=-1, encoding=None, errors=None, newline=None, line_buffering=False, bufsize=-1, **kwargs):
        npath = self._normpath(path)
        if self.isdir(path):
            msg = "that's a directory: %(path)s"
            raise ResourceInvalidError(path, msg=msg)
        #  paramiko implements its own buffering and write-back logic,
        #  so we don't need to use a RemoteFileBuffer here.
        f = self.client.open(npath, mode, bufsize)
        #  Unfortunately it has a broken truncate() method.
        #  TODO: implement this as a wrapper
        old_truncate = f.truncate

        def new_truncate(size=None):
            if size is None:
                size = f.tell()
            return old_truncate(size)
        f.truncate = new_truncate
        return f

    @synchronize
    def desc(self, path):
        npath = self._normpath(path)
        if self.hostname:
            return u'sftp://%s%s' % (self.hostname, path)
        else:
            addr, port = self._transport.getpeername()
            return u'sftp://%s:%i%s' % (addr, port, self.client.normalize(npath))

    @synchronize
    @convert_os_errors
    def exists(self, path):
        if path in ('', '/'):
            return True
        npath = self._normpath(path)
        try:
            self.client.stat(npath)
        except IOError, e:
            if getattr(e,"errno",None) == ENOENT:
                return False
            raise
        return True

    @synchronize
    @convert_os_errors
    def isdir(self,path):
        if normpath(path) in ('', '/'):
            return True
        npath = self._normpath(path)
        try:
            stat = self.client.stat(npath)
        except IOError, e:
            if getattr(e,"errno",None) == ENOENT:
                return False
            raise
        return statinfo.S_ISDIR(stat.st_mode) != 0

    @synchronize
    @convert_os_errors
    def isfile(self,path):
        npath = self._normpath(path)
        try:
            stat = self.client.stat(npath)
        except IOError, e:
            if getattr(e,"errno",None) == ENOENT:
                return False
            raise
        return statinfo.S_ISREG(stat.st_mode) != 0

    @synchronize
    @convert_os_errors
    def listdir(self,path="./",wildcard=None,full=False,absolute=False,dirs_only=False,files_only=False):
        npath = self._normpath(path)
        try:
            attrs_map = None
            if dirs_only or files_only:
                attrs = self.client.listdir_attr(npath)
                attrs_map = dict((a.filename, a) for a in attrs)
                paths = list(attrs_map.iterkeys())
            else:
                paths = self.client.listdir(npath)
        except IOError, e:
            if getattr(e,"errno",None) == ENOENT:
                if self.isfile(path):
                    raise ResourceInvalidError(path,msg="Can't list directory contents of a file: %(path)s")
                raise ResourceNotFoundError(path)
            elif self.isfile(path):
                raise ResourceInvalidError(path,msg="Can't list directory contents of a file: %(path)s")
            raise

        if attrs_map:
            if dirs_only:
                filter_paths = []
                for apath, attr in attrs_map.iteritems():
                    if isdir(self, path, attr.__dict__):
                        filter_paths.append(apath)
                paths = filter_paths
            elif files_only:
                filter_paths = []
                for apath, attr in attrs_map.iteritems():
                    if isfile(self, apath, attr.__dict__):
                        filter_paths.append(apath)
                paths = filter_paths

        for (i,p) in enumerate(paths):
            if not isinstance(p,unicode):
                paths[i] = p.decode(self.encoding)

        return self._listdir_helper(path, paths, wildcard, full, absolute, False, False)

    @synchronize
    @convert_os_errors
    def listdirinfo(self,path="./",wildcard=None,full=False,absolute=False,dirs_only=False,files_only=False):
        npath = self._normpath(path)
        try:
            attrs = self.client.listdir_attr(npath)
            attrs_map = dict((a.filename, a) for a in attrs)
            paths = attrs_map.keys()
        except IOError, e:
            if getattr(e,"errno",None) == ENOENT:
                if self.isfile(path):
                    raise ResourceInvalidError(path,msg="Can't list directory contents of a file: %(path)s")
                raise ResourceNotFoundError(path)
            elif self.isfile(path):
                raise ResourceInvalidError(path,msg="Can't list directory contents of a file: %(path)s")
            raise

        if dirs_only:
            filter_paths = []
            for path, attr in attrs_map.iteritems():
                if isdir(self, path, attr.__dict__):
                    filter_paths.append(path)
            paths = filter_paths
        elif files_only:
            filter_paths = []
            for path, attr in attrs_map.iteritems():
                if isfile(self, path, attr.__dict__):
                    filter_paths.append(path)
            paths = filter_paths

        for (i, p) in enumerate(paths):
            if not isinstance(p, unicode):
                paths[i] = p.decode(self.encoding)

        def getinfo(p):
            resourcename = basename(p)
            info = attrs_map.get(resourcename)
            if info is None:
                return self.getinfo(pathjoin(path, p))
            return self._extract_info(info.__dict__)

        return [(p, getinfo(p)) for p in
                    self._listdir_helper(path, paths, wildcard, full, absolute, False, False)]

    @synchronize
    @convert_os_errors
    def makedir(self,path,recursive=False,allow_recreate=False):
        npath = self._normpath(path)
        try:
            self.client.mkdir(npath)
        except IOError, _e:
            # Error code is unreliable, try to figure out what went wrong
            try:
                stat = self.client.stat(npath)
            except IOError:
                if not self.isdir(dirname(path)):
                    # Parent dir is missing
                    if not recursive:
                        raise ParentDirectoryMissingError(path)
                    self.makedir(dirname(path),recursive=True)
                    self.makedir(path,allow_recreate=allow_recreate)
                else:
                    # Undetermined error, let the decorator handle it
                    raise
            else:
                # Destination exists
                if statinfo.S_ISDIR(stat.st_mode):
                    if not allow_recreate:
                        raise DestinationExistsError(path,msg="Can't create a directory that already exists (try allow_recreate=True): %(path)s")
                else:
                    raise ResourceInvalidError(path,msg="Can't create directory, there's already a file of that name: %(path)s")

    @synchronize
    @convert_os_errors
    def remove(self,path):
        npath = self._normpath(path)
        try:
            self.client.remove(npath)
        except IOError, e:
            if getattr(e,"errno",None) == ENOENT:
                raise ResourceNotFoundError(path)
            elif self.isdir(path):
                raise ResourceInvalidError(path,msg="Cannot use remove() on a directory: %(path)s")
            raise

    @synchronize
    @convert_os_errors
    def removedir(self,path,recursive=False,force=False):
        npath = self._normpath(path)
        if normpath(path) in ('', '/'):
            raise RemoveRootError(path)
        if force:
            for path2 in self.listdir(path,absolute=True):
                try:
                    self.remove(path2)
                except ResourceInvalidError:
                    self.removedir(path2,force=True)
        if not self.exists(path):
            raise ResourceNotFoundError(path)
        try:
            self.client.rmdir(npath)
        except IOError, e:
            if getattr(e,"errno",None) == ENOENT:
                if self.isfile(path):
                    raise ResourceInvalidError(path,msg="Can't use removedir() on a file: %(path)s")
                raise ResourceNotFoundError(path)

            elif self.listdir(path):
                raise DirectoryNotEmptyError(path)
            raise
        if recursive:
            try:
                if dirname(path) not in ('', '/'):
                    self.removedir(dirname(path),recursive=True)
            except DirectoryNotEmptyError:
                pass

    @synchronize
    @convert_os_errors
    def rename(self,src,dst):
        nsrc = self._normpath(src)
        ndst = self._normpath(dst)
        try:
            self.client.rename(nsrc,ndst)
        except IOError, e:
            if getattr(e,"errno",None) == ENOENT:
                raise ResourceNotFoundError(src)
            if not self.isdir(dirname(dst)):
                raise ParentDirectoryMissingError(dst)
            raise

    @synchronize
    @convert_os_errors
    def move(self,src,dst,overwrite=False,chunk_size=16384):
        nsrc = self._normpath(src)
        ndst = self._normpath(dst)
        if overwrite and self.isfile(dst):
            self.remove(dst)
        try:
            self.client.rename(nsrc,ndst)
        except IOError, e:
            if getattr(e,"errno",None) == ENOENT:
                raise ResourceNotFoundError(src)
            if self.exists(dst):
                raise DestinationExistsError(dst)
            if not self.isdir(dirname(dst)):
                raise ParentDirectoryMissingError(dst,msg="Destination directory does not exist: %(path)s")
            raise

    @synchronize
    @convert_os_errors
    def movedir(self,src,dst,overwrite=False,ignore_errors=False,chunk_size=16384):
        nsrc = self._normpath(src)
        ndst = self._normpath(dst)
        if overwrite and self.isdir(dst):
            self.removedir(dst)
        try:
            self.client.rename(nsrc,ndst)
        except IOError, e:
            if getattr(e,"errno",None) == ENOENT:
                raise ResourceNotFoundError(src)
            if self.exists(dst):
                raise DestinationExistsError(dst)
            if not self.isdir(dirname(dst)):
                raise ParentDirectoryMissingError(dst,msg="Destination directory does not exist: %(path)s")
            raise

    _info_vars = frozenset('st_size st_uid st_gid st_mode st_atime st_mtime'.split())
    @classmethod
    def _extract_info(cls, stats):
        fromtimestamp = datetime.datetime.fromtimestamp
        info = dict((k, v) for k, v in stats.iteritems() if k in cls._info_vars and not k.startswith('_'))
        info['size'] = info['st_size']
        ct = info.get('st_ctime')
        if ct is not None:
            info['created_time'] = fromtimestamp(ct)
        at = info.get('st_atime')
        if at is not None:
            info['accessed_time'] = fromtimestamp(at)
        mt = info.get('st_mtime')
        if mt is not None:
            info['modified_time'] = fromtimestamp(mt)
        return info

    @synchronize
    @convert_os_errors
    def getinfo(self, path):
        npath = self._normpath(path)
        stats = self.client.stat(npath)
        info = dict((k, getattr(stats, k)) for k in dir(stats) if not k.startswith('_'))
        info['size'] = info['st_size']
        ct = info.get('st_ctime', None)
        if ct is not None:
            info['created_time'] = datetime.datetime.fromtimestamp(ct)
        at = info.get('st_atime', None)
        if at is not None:
            info['accessed_time'] = datetime.datetime.fromtimestamp(at)
        mt = info.get('st_mtime', None)
        if mt is not None:
            info['modified_time'] = datetime.datetime.fromtimestamp(mt)
        return info

    @synchronize
    @convert_os_errors
    def getsize(self, path):
        npath = self._normpath(path)
        stats = self.client.stat(npath)
        return stats.st_size



#-------------------------------------------------------------------------
#
#  Copyright (c) 2007, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   08/11/2007
#
#-------------------------------------------------------------------------

""" Defines a themed panel that wraps itself around a single child widget.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

import wx

from traits.api \
    import Str, Property, Instance, Bool, cached_property

from themed_window \
    import ThemedWindow

#-------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------

# Size of an empty text string:
ZeroTextSize = (0, 0, 0, 0)

#-------------------------------------------------------------------------
#  'ImagePanel' class:
#-------------------------------------------------------------------------


class ImagePanel(ThemedWindow):

    # The optional text to display in the top or bottom of the image slice:
    text = Str(event='updated')

    # Can the application change the theme contents?
    mutable_theme = Bool(False)

    # Is the image panel capable of displaying text?
    can_show_text = Property

    # The adjusted size of the panel, taking into account the size of its
    # current children and the image border:
    adjusted_size = Property

    # The best size of the panel, taking into account the best size of its
    # children and the image border:
    best_size = Property

    # The underlying wx control:
    control = Instance(wx.Window)

    #-- Private Traits -------------------------------------------------------

    # The size of the current text:
    text_size = Property(depends_on='text, control')

    #-- Public Methods -------------------------------------------------------

    def create_control(self, parent):
        """ Creates the underlying wx.Panel control.
        """
        self.control = control = wx.Panel(
            parent, -1, style=wx.TAB_TRAVERSAL | wx.FULL_REPAINT_ON_RESIZE)

        # Set up the sizer for the control:
        control.SetSizer(ImageSizer(self.theme))

        # Initialize the control (set-up event handlers, ...)
        self.init_control()

        # Attach the image slice to the control:
        control._image_slice = self.theme.image_slice

        # Set the panel's background colour to the image slice bg_color:
        control.SetBackgroundColour(control._image_slice.bg_color)

        return control

    def layout(self):
        """ Lays out the contents of the image panel.
        """
        self.control.Layout()
        self.control.Refresh()

    #-- Property Implementations ---------------------------------------------

    def _get_adjusted_size(self):
        """ Returns the adjusted size of the panel taking into account the
            size of its current children and the image border.
        """
        control = self.control
        dx, dy = 0, 0
        for child in control.GetChildren():
            dx, dy = child.GetSizeTuple()

        size = self._adjusted_size_of(dx, dy)
        control.SetSize(size)

        return size

    def _get_best_size(self):
        """ Returns the best size of the panel taking into account the
            best size of its current children and the image border.
        """
        control = self.control
        dx, dy = 0, 0
        for child in control.GetChildren():
            dx, dy = child.GetBestSize()

        return self._adjusted_size_of(dx, dy)

    @cached_property
    def _get_can_show_text(self):
        """ Returns whether or not the image panel is capable of displaying
            text.
        """
        tdx, tdy, descent, leading = self.control.GetFullTextExtent('Myj')
        slice = self.theme.image_slice
        tdy += 4
        return ((tdy <= slice.xtop) or (tdy <= slice.xbottom) or
                (slice.xleft >= 40) or (slice.xright >= 40))

    @cached_property
    def _get_text_size(self):
        """ Returns the text size information for the window.
        """
        if (self.text == '') or (self.control is None):
            return ZeroTextSize

        return self.control.GetFullTextExtent(self.text)

    #-- Trait Event Handlers -------------------------------------------------

    def _updated_changed(self):
        """ Handles a change that requires the control to be updated.
        """
        if self.control is not None:
            self.control.Refresh()

    def _mutable_theme_changed(self, state):
        """ Handles a change to the 'mutable_theme' trait.
        """
        self.on_trait_change(
            self._theme_modified,
            'theme.[border.-,content.-,label.-,alignment,content_color,'
            'label_color]',
            remove=not state)

    def _theme_modified(self):
        if self.control is not None:
            self.layout()

    def _theme_changed(self, theme):
        """ Handles the 'theme' trait being changed.
        """
        super(ImagePanel, self)._theme_changed()

        control = self.control
        if (control is not None) and (theme is not None):
            # Attach the image slice to the control:
            control._image_slice = theme.image_slice

            # Set the panel's background colour to the image slice bg_color:
            control.SetBackgroundColour(control._image_slice.bg_color)

    #-- wx.Python Event Handlers ---------------------------------------------

    def _paint_fg(self, dc):
        """ Paints the foreground into the specified device context.
        """
        # If we have text and have room to draw it, then do so:
        text = self.text
        if (text != '') and self.can_show_text:
            theme = self.theme
            dc.SetBackgroundMode(wx.TRANSPARENT)
            dc.SetTextForeground(theme.label_color)
            dc.SetFont(self.control.GetFont())

            alignment = theme.alignment
            label = theme.label
            wdx, wdy = self.control.GetClientSizeTuple()
            tdx, tdy, descent, leading = self.text_size
            tx = None
            slice = theme.image_slice
            xleft = slice.xleft
            xright = slice.xright
            xtop = slice.xtop
            xbottom = slice.xbottom
            ltop = label.top
            lbottom = label.bottom
            tdyp = tdy + ltop + lbottom
            cl = xleft + label.left
            cr = wdx - xright - label.right
            if (tdyp <= xtop) and (xtop >= xbottom):
                ty = ((ltop + xtop - lbottom - tdy) / 2) + 1
            elif tdy <= xbottom:
                ty = wdy + ((ltop - xbottom - lbottom - tdy) / 2)
            else:
                ty = (wdy + xtop + label.top - xbottom - label.bottom - tdy) / 2
                if xleft >= xright:
                    cl = label.left
                    cr = xleft - label.right
                else:
                    cl = wdx - xright + label.left
                    cr = wdx - label.right

            # Calculate the x coordinate for the specified alignment type:
            if alignment == 'left':
                tx = cl
            elif alignment == 'right':
                tx = cr - tdx
            else:
                tx = (cl + cr - tdx) / 2

            # Draw the (clipped) text string:
            dc.SetClippingRegion(cl, ty, cr - cl, tdy)
            dc.DrawText(text, tx, ty)
            dc.DestroyClippingRegion()

    #-- Private Methods ------------------------------------------------------

    def _adjusted_size_of(self, dx, dy):
        """ Returns the adjusted size of its children, taking into account the
            image slice border.
        """
        slice = self.theme.image_slice
        content = self.theme.content
        sizer = self.control.GetSizer()
        return wx.Size(dx + min(slice.left, slice.xleft) +
                       min(slice.right, slice.xright) +
                       content.left + content.right,
                       dy + min(slice.top, slice.xtop) +
                       min(slice.bottom, slice.xbottom) +
                       content.top + content.bottom)

#-------------------------------------------------------------------------
#  'ImageSizer' class:
#-------------------------------------------------------------------------


class ImageSizer(wx.PySizer):
    """ Defines a sizer that correctly sizes a window's children to fit within
        the borders implicitly defined by a background ImageSlice object,
    """

    #-------------------------------------------------------------------------
    #  Initializes the object:
    #-------------------------------------------------------------------------

    def __init__(self, theme):
        """ Initializes the object.
        """
        super(ImageSizer, self).__init__()

        # Save a reference to the theme:
        self._theme = theme

        # Save the ImageSlice object which determines the inset border size:
        self._image_slice = theme.image_slice

    #-------------------------------------------------------------------------
    #  Calculates the minimum size needed by the sizer:
    #-------------------------------------------------------------------------

    def CalcMin(self):
        """ Calculates the minimum size of the control by adding its contents
            minimum size to the ImageSlice object's border size.
        """
        dx, dy = 0, 0
        for item in self.GetChildren():
            if item.IsSizer():
                dx, dy = item.GetSizer().CalcMin()
            else:
                dx, dy = item.GetWindow().GetBestSize()

        slice = self._image_slice
        content = self._theme.content

        return wx.Size(max(slice.left + slice.right,
                           slice.xleft + slice.xright +
                           content.left + content.right + dx),
                       max(slice.top + slice.bottom,
                           slice.xtop + slice.xbottom +
                           content.top + content.bottom + dy))

    #-------------------------------------------------------------------------
    #  Layout the contents of the sizer based on the sizer's current size and
    #  position:
    #-------------------------------------------------------------------------

    def RecalcSizes(self):
        """ Layout the contents of the sizer based on the sizer's current size
            and position.
        """
        x, y = self.GetPositionTuple()
        dx, dy = self.GetSizeTuple()
        slice = self._image_slice
        content = self._theme.content
        left = slice.xleft + content.left
        top = slice.xtop + content.top
        ix, iy, idx, idy = (x + left,
                            y + top,
                            dx - left - slice.xright - content.right,
                            dy - top - slice.xbottom - content.bottom)

        for item in self.GetChildren():
            if item.IsSizer():
                item.GetSizer().SetDimension(ix, iy, idx, idy)
            else:
                item.GetWindow().SetDimensions(ix, iy, idx, idy)



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/21/2004
#
#------------------------------------------------------------------------------

""" Defines the various Boolean editors for the wxPython user interface toolkit.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

import wx

# FIXME: ToolkitEditorFactory is a proxy class defined here just for backward
# compatibility. The class has been moved to the
# traitsui.editors.boolean_editor file.
from traitsui.editors.boolean_editor \
    import ToolkitEditorFactory

from editor \
    import Editor

# This needs to be imported in here for use by the editor factory for boolean
# editors (declared in traitsui). The editor factory's text_editor
# method will use the TextEditor in the ui.
from text_editor \
    import SimpleEditor as TextEditor

from constants \
    import ReadonlyColor

#-------------------------------------------------------------------------
#  'SimpleEditor' class:
#-------------------------------------------------------------------------


class SimpleEditor(Editor):
    """ Simple style of editor for Boolean values, which displays a check box.
    """
    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        self.control = wx.CheckBox(parent, -1, '')
        wx.EVT_CHECKBOX(parent, self.control.GetId(), self.update_object)
        self.set_tooltip()

    #-------------------------------------------------------------------------
    #  Handles the user clicking on the checkbox:
    #-------------------------------------------------------------------------

    def update_object(self, event):
        """ Handles the user clicking the checkbox.
        """
        self.value = (self.control.GetValue() != 0)

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        self.control.SetValue(self.value)

#-------------------------------------------------------------------------
#  'ReadonlyEditor' class:
#-------------------------------------------------------------------------


class ReadonlyEditor(Editor):
    """ Read-only style of editor for Boolean values, which displays static text
    of either "True" or "False".
    """
    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        self.control = wx.TextCtrl(parent, -1, '', style=wx.TE_READONLY)
        self.control.SetBackgroundColour(ReadonlyColor)

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #
    #  (Should normally be overridden in a subclass)
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        if self.value:
            self.control.SetValue('True')
        else:
            self.control.SetValue('False')



#------------------------------------------------------------------------------
#
#  Copyright (c) 2012, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Author: Pietro Berkes
#  Date:   Jan 2012
#
#------------------------------------------------------------------------------

"""
Test case for bug (wx, Mac OS X)

A RangeEditor in mode 'text' for an Int allows values out of range.
"""

from __future__ import print_function

from traits.has_traits import HasTraits
from traits.trait_types import Int
from traitsui.item import Item
from traitsui.view import View
from traitsui.editors.range_editor import RangeEditor

from traitsui.tests._tools import *


class NumberWithTextEditor(HasTraits):
    """Dialog containing a RangeEditor in 'spinner' mode for an Int.
    """

    number = Int

    traits_view = View(
        Item(label="Range should be 3 to 8. Enter 1, then press OK"),
        Item('number', editor=RangeEditor(low=3, high=8, mode='text')),
        buttons=['OK']
    )


@skip_if_not_wx
def test_wx_spin_control_editing():
    # behavior: when editing the text part of a spin control box, pressing
    # the OK button should update the value of the HasTraits class
    # (tests a bug where this fails with an AttributeError)

    with store_exceptions_on_all_threads():
        num = NumberWithTextEditor()
        ui = num.edit_traits()

        # the following is equivalent to setting the text in the text control,
        # then pressing OK

        textctrl = ui.control.FindWindowByName('text')
        textctrl.SetValue('1')

        # press the OK button and close the dialog
        press_ok_button(ui)

    # the number traits should be between 3 and 8
    assert num.number >= 3 and num.number <= 8


if __name__ == '__main__':
    # Executing the file opens the dialog for manual testing
    num = NumberWithTextEditor()
    num.configure_traits()
    print(num.number)



#!/usr/bin/env python

"""
Testing of code to find nodes.

Currently only nearest neighbor.

"""

from __future__ import (absolute_import, division, print_function)

import numpy as np

from .utilities import twenty_one_triangles


def test_locate_node():
    """Test finding a single node."""

    ugrid = twenty_one_triangles()
    assert ugrid.locate_nodes((4.58, 5.08)) == 6


def test_locate_nodes():
    """Test finding multiple nodes at once."""

    ugrid = twenty_one_triangles()
    assert np.array_equal(ugrid.locate_nodes(((4.58, 5.08),
                                              (4.81, 0.89),
                                              (6.43, 12.9),
                                              (8.74, 6.86),
                                              (5.12, 7.31),
                                              )), (6, 0, 17, 10, 8))


def test_locate_exact():
    """
    The nearest neighbor of the exact node locations had better be
    the nodes!

    """
    ugrid = twenty_one_triangles()
    assert np.array_equal(ugrid.locate_nodes(ugrid.nodes),
                          list(range(len(ugrid.nodes))))


def test_locate_middle():
    """See what happens the point is equidistant to two nodes."""
    ugrid = twenty_one_triangles()

    # (3,5) is equidistant between nodes 2, 6 and 7
    # 2 is returned, but might be arbitrary
    # assert ugrid.locate_nodes( (3, 5) ) == 2

    # Perturb the point a bit, and nearest changes to:
    assert ugrid.locate_nodes((3.0000000001, 5)) == 6
    assert ugrid.locate_nodes((3, 5.00000000001)) == 7
    assert ugrid.locate_nodes((3, 4.99999999999)) == 2


if __name__ == '__main__':
    test_locate_nodes()



#  Copyright (c) 2009-2010, Cloud Matrix Pty. Ltd.
#  All rights reserved; available under the terms of the MIT License.
"""

  fs.expose.dokan.libdokan:  low-level ctypes interface to Dokan

"""

from ctypes import *

try:
    DokanMain = windll.Dokan.DokanMain
    DokanVersion = windll.Dokan.DokanVersion
except AttributeError:
    raise ImportError("Dokan DLL not found")


from ctypes.wintypes import *
ULONG64 = c_ulonglong
ULONGLONG = c_ulonglong
PULONGLONG = POINTER(ULONGLONG)
UCHAR = c_ubyte
LPDWORD = POINTER(DWORD)
LONGLONG = c_longlong

try:
    USHORT = USHORT
except NameError:
    #  Not available in older python versions
    USHORT = c_ushort


DokanVersion.restype = ULONG
DokanVersion.argtypes = ()
if DokanVersion() < 392:  # ths is release 0.5.3
    raise ImportError("Dokan DLL is too old")


MAX_PATH = 260

class FILETIME(Structure):
    _fields_ = [
        ("dwLowDateTime", DWORD),
        ("dwHighDateTime", DWORD),
    ]

class WIN32_FIND_DATAW(Structure):
    _fields_ = [
        ("dwFileAttributes", DWORD),
        ("ftCreationTime", FILETIME),
        ("ftLastAccessTime", FILETIME),
        ("ftLastWriteTime", FILETIME),
        ("nFileSizeHigh", DWORD),
        ("nFileSizeLow", DWORD),
        ("dwReserved0", DWORD),
        ("dwReserved1", DWORD),
        ("cFileName", WCHAR * MAX_PATH),
        ("cAlternateFileName", WCHAR * 14),
    ]

class BY_HANDLE_FILE_INFORMATION(Structure):
    _fields_ = [
        ('dwFileAttributes', DWORD),
        ('ftCreationTime', FILETIME),
        ('ftLastAccessTime', FILETIME),
        ('ftLastWriteTime', FILETIME),
        ('dwVolumeSerialNumber', DWORD),
        ('nFileSizeHigh', DWORD),
        ('nFileSizeLow', DWORD),
        ('nNumberOfLinks', DWORD),
        ('nFileIndexHigh', DWORD),
        ('nFileIndexLow', DWORD),
    ]

class DOKAN_OPTIONS(Structure):
    _fields_ = [
	("DriveLetter", WCHAR),
	("ThreadCount", USHORT),
	("Options", ULONG),
	("GlobalContext", ULONG64),
    ]


class DOKAN_FILE_INFO(Structure):
    _fields_ = [
        ("Context", ULONG64),
        ("DokanContext", ULONG64),
        ("DokanOptions", POINTER(DOKAN_OPTIONS)),
        ("ProcessId", ULONG),
        ("IsDirectory", UCHAR),
        ("DeleteOnClose", UCHAR),
        ("PagingIO", UCHAR),
        ("SyncronousIo", UCHAR),
        ("Nocache", UCHAR),
        ("WriteToEndOfFile", UCHAR),
]


PDOKAN_FILE_INFO = POINTER(DOKAN_FILE_INFO)
PFillFindData = WINFUNCTYPE(c_int,POINTER(WIN32_FIND_DATAW),PDOKAN_FILE_INFO)

class DOKAN_OPERATIONS(Structure):
    _fields_ = [
        ("CreateFile", CFUNCTYPE(c_int,
                LPCWSTR,      # FileName
                DWORD,        # DesiredAccess
                DWORD,        # ShareMode
                DWORD,        # CreationDisposition
                DWORD,        # FlagsAndAttributes
                PDOKAN_FILE_INFO)),
        ("OpenDirectory", CFUNCTYPE(c_int,
		LPCWSTR,      # FileName
		PDOKAN_FILE_INFO)),
        ("CreateDirectory", CFUNCTYPE(c_int,
		LPCWSTR,      # FileName
		PDOKAN_FILE_INFO)),
        ("Cleanup", CFUNCTYPE(c_int,
		LPCWSTR,      # FileName
		PDOKAN_FILE_INFO)),
        ("CloseFile", CFUNCTYPE(c_int,
		LPCWSTR,      # FileName
		PDOKAN_FILE_INFO)),
        ("ReadFile", CFUNCTYPE(c_int,
		LPCWSTR,  # FileName
		POINTER(c_char),   # Buffer
		DWORD,    # NumberOfBytesToRead
		LPDWORD,  # NumberOfBytesRead
		LONGLONG, # Offset
		PDOKAN_FILE_INFO)),
        ("WriteFile", CFUNCTYPE(c_int,
		LPCWSTR,  # FileName
		POINTER(c_char), # Buffer
		DWORD,    # NumberOfBytesToWrite
		LPDWORD,  # NumberOfBytesWritten
		LONGLONG, # Offset
		PDOKAN_FILE_INFO)),
        ("FlushFileBuffers", CFUNCTYPE(c_int,
		LPCWSTR,  # FileName
		PDOKAN_FILE_INFO)),
        ("GetFileInformation", CFUNCTYPE(c_int,
		LPCWSTR,  # FileName
		POINTER(BY_HANDLE_FILE_INFORMATION), # Buffer
		PDOKAN_FILE_INFO)),
        ("FindFiles", CFUNCTYPE(c_int,
		LPCWSTR,  # PathName
		PFillFindData, # call this function with PWIN32_FIND_DATAW
		PDOKAN_FILE_INFO)),
        ("FindFilesWithPattern", CFUNCTYPE(c_int,
		LPCWSTR,  # PathName
		LPCWSTR,  # SearchPattern
		PFillFindData,	#call this function with PWIN32_FIND_DATAW
		PDOKAN_FILE_INFO)),
        ("SetFileAttributes", CFUNCTYPE(c_int,
		LPCWSTR, # FileName
		DWORD,   # FileAttributes
		PDOKAN_FILE_INFO)),
        ("SetFileTime", CFUNCTYPE(c_int,
		LPCWSTR, # FileName
		POINTER(FILETIME), # CreationTime
		POINTER(FILETIME), # LastAccessTime
		POINTER(FILETIME), # LastWriteTime
		PDOKAN_FILE_INFO)),
        ("DeleteFile", CFUNCTYPE(c_int,
		LPCWSTR, # FileName
		PDOKAN_FILE_INFO)),
        ("DeleteDirectory", CFUNCTYPE(c_int,
		LPCWSTR, # FileName
		PDOKAN_FILE_INFO)),
        ("MoveFile", CFUNCTYPE(c_int,
		LPCWSTR, # ExistingFileName
		LPCWSTR, # NewFileName
		BOOL,	 # ReplaceExisiting
		PDOKAN_FILE_INFO)),
        ("SetEndOfFile", CFUNCTYPE(c_int,
		LPCWSTR,  # FileName
		LONGLONG, # Length
		PDOKAN_FILE_INFO)),
        ("SetAllocationSize", CFUNCTYPE(c_int,
		LPCWSTR,  # FileName
		LONGLONG, # Length
		PDOKAN_FILE_INFO)),
        ("LockFile", CFUNCTYPE(c_int,
		LPCWSTR,  # FileName
		LONGLONG, # ByteOffset
		LONGLONG, # Length
		PDOKAN_FILE_INFO)),
        ("UnlockFile", CFUNCTYPE(c_int,
		LPCWSTR,  # FileName
		LONGLONG, # ByteOffset
		LONGLONG, # Length
		PDOKAN_FILE_INFO)),
        ("GetDiskFreeSpaceEx", CFUNCTYPE(c_int,
		PULONGLONG, # FreeBytesAvailable
		PULONGLONG, # TotalNumberOfBytes
		PULONGLONG, # TotalNumberOfFreeBytes
		PDOKAN_FILE_INFO)),
        ("GetVolumeInformation", CFUNCTYPE(c_int,
		POINTER(c_wchar),  # VolumeNameBuffer
		DWORD,	 # VolumeNameSize in num of chars
		LPDWORD, # VolumeSerialNumber
		LPDWORD, # MaximumComponentLength in num of chars
		LPDWORD, # FileSystemFlags
		POINTER(c_wchar),  # FileSystemNameBuffer
		DWORD,	 # FileSystemNameSize in num of chars
		PDOKAN_FILE_INFO)),
        ("Unmount", CFUNCTYPE(c_int,
		PDOKAN_FILE_INFO)),
    ]



DokanMain.restype = c_int
DokanMain.argtypes = (
    POINTER(DOKAN_OPTIONS),
    POINTER(DOKAN_OPERATIONS),
)



DokanUnmount = windll.Dokan.DokanUnmount
DokanUnmount.restype = BOOL
DokanUnmount.argtypes = (
    WCHAR,
)

DokanIsNameInExpression = windll.Dokan.DokanIsNameInExpression
DokanIsNameInExpression.restype = BOOL
DokanIsNameInExpression.argtypes = (
    LPCWSTR,  # pattern
    LPCWSTR,  # name
    BOOL,     # ignore case
)

DokanDriverVersion = windll.Dokan.DokanDriverVersion
DokanDriverVersion.restype = ULONG
DokanDriverVersion.argtypes = (
)

DokanResetTimeout = windll.Dokan.DokanResetTimeout
DokanResetTimeout.restype = BOOL
DokanResetTimeout.argtypes = (
    ULONG,  #timeout
    PDOKAN_FILE_INFO,  # file info pointer
)





#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/21/2004
#
#------------------------------------------------------------------------------

""" Defines the various font editors and the font editor factory, for the
    wxPython user interface toolkit..
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

import wx

from traits.api import Bool

from traitsui.editors.font_editor \
    import ToolkitEditorFactory as BaseToolkitEditorFactory

from editor_factory \
    import SimpleEditor as BaseSimpleEditor, \
    TextEditor as BaseTextEditor, \
    ReadonlyEditor as BaseReadonlyEditor

from editor \
    import Editor

from helper \
    import TraitsUIPanel, disconnect

#-------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------

# Standard font point sizes
PointSizes = [
    '8', '9', '10', '11', '12', '14', '16', '18',
    '20', '22', '24', '26', '28', '36', '48', '72'
]

# All available font styles
Styles = ['Normal', 'Slant', 'Italic']

# All available font weights
Weights = ['Normal', 'Light', 'Bold']

# All available font facenames
facenames = None

#---------------------------------------------------------------------------
#  The wxPython ToolkitEditorFactory class.
#---------------------------------------------------------------------------
## We need to add wx-specific methods to the editor factory, and so we create
## a subclass of the BaseToolkitEditorFactory.


class ToolkitEditorFactory(BaseToolkitEditorFactory):
    """ wxPython editor factory for font editors.
    """

    show_style = Bool(False)
    show_weight = Bool(False)

    #-------------------------------------------------------------------------
    #  Returns a wxFont object corresponding to a specified object's font trait:
    #-------------------------------------------------------------------------

    def to_wx_font(self, editor):
        """ Returns a wxFont object corresponding to a specified object's font
        trait.
        """
        font = editor.value
        return wx.Font(font.GetPointSize(), font.GetFamily(), font.GetStyle(),
                       font.GetWeight(), font.GetUnderlined(),
                       font.GetFaceName())

    #-------------------------------------------------------------------------
    #  Gets the application equivalent of a wxPython Font value:
    #-------------------------------------------------------------------------

    def from_wx_font(self, font):
        """ Gets the application equivalent of a wxPython Font value.
        """
        return font

    #-------------------------------------------------------------------------
    #  Returns the text representation of the specified object trait value:
    #-------------------------------------------------------------------------

    def str_font(self, font):
        """ Returns the text representation of the specified object trait value.
        """
        weight = {wx.LIGHT: ' Light',
                  wx.BOLD: ' Bold'}.get(font.GetWeight(), '')
        style = {wx.SLANT: ' Slant',
                 wx.ITALIC: ' Italic'}.get(font.GetStyle(), '')
        return '%s point %s%s%s' % (
            font.GetPointSize(), font.GetFaceName(), style, weight)

    #-------------------------------------------------------------------------
    #  Returns a list of all available font facenames:
    #-------------------------------------------------------------------------

    def all_facenames(self):
        """ Returns a list of all available font facenames.
        """
        global facenames

        if facenames is None:
            facenames = sorted(FontEnumerator().facenames())
        return facenames

#-------------------------------------------------------------------------
#  'SimpleFontEditor' class:
#-------------------------------------------------------------------------


class SimpleFontEditor(BaseSimpleEditor):
    """ Simple style of font editor, which displays a text field that contains
        a text representation of the font value (using that font if possible).
        Clicking the field displays a font selection dialog box.
    """

    #-------------------------------------------------------------------------
    #  Invokes the pop-up editor for an object trait:
    #-------------------------------------------------------------------------

    def popup_editor(self, event):
        """ Invokes the pop-up editor for an object trait.
        """
        font_data = wx.FontData()
        font_data.SetInitialFont(self.factory.to_wx_font(self))
        dialog = wx.FontDialog(self.control, font_data)
        if dialog.ShowModal() == wx.ID_OK:
            self.value = self.factory.from_wx_font(
                dialog.GetFontData().GetChosenFont())
            self.update_editor()

        dialog.Destroy()

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        super(SimpleFontEditor, self).update_editor()
        set_font(self)

    #-------------------------------------------------------------------------
    #  Returns the text representation of a specified font value:
    #-------------------------------------------------------------------------

    def string_value(self, font):
        """ Returns the text representation of a specified font value.
        """
        return self.factory.str_font(font)

#-------------------------------------------------------------------------
#  'CustomFontEditor' class:
#-------------------------------------------------------------------------


class CustomFontEditor(Editor):
    """ Custom style of font editor, which displays the following:

        * A combo box containing all available type face names.
        * A combo box containing the available type sizes.
        * A combo box containing the available type styles
    """

    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        # Create a panel to hold all of the buttons:
        self.control = panel = TraitsUIPanel(parent, -1)
        sizer = wx.BoxSizer(wx.VERTICAL)

        # Add all of the font choice controls:
        sizer2 = wx.BoxSizer(wx.HORIZONTAL)
        facenames = self.factory.all_facenames()
        control = self._facename = wx.Choice(panel, -1, wx.Point(0, 0),
                                             wx.Size(-1, -1), facenames)

        sizer2.Add(control, 4, wx.EXPAND)
        wx.EVT_CHOICE(panel, control.GetId(), self.update_object_parts)

        control = self._point_size = wx.Choice(panel, -1, wx.Point(0, 0),
                                               wx.Size(-1, -1), PointSizes)
        sizer2.Add(control, 1, wx.EXPAND | wx.LEFT, 3)
        wx.EVT_CHOICE(panel, control.GetId(), self.update_object_parts)

        if self.factory.show_style:
            self._style = wx.Choice(panel, -1, wx.Point(0, 0),
                                    wx.Size(-1, -1), Styles)
            sizer2.Add(self._style, 1, wx.EXPAND | wx.LEFT, 3)
            wx.EVT_CHOICE(panel, self._style.GetId(), self.update_object_parts)

        if self.factory.show_weight:
            self._weight = wx.Choice(panel, -1, wx.Point(0, 0),
                                     wx.Size(-1, -1), Weights)
            sizer2.Add(self._weight, 1, wx.EXPAND | wx.LEFT, 3)
            wx.EVT_CHOICE(
                panel,
                self._weight.GetId(),
                self.update_object_parts)

        sizer.Add(sizer2, 0, wx.EXPAND)

        # Set-up the layout:
        panel.SetSizer(sizer)

        self.set_tooltip()

    #-------------------------------------------------------------------------
    #  Disposes of the contents of an editor:
    #-------------------------------------------------------------------------

    def dispose(self):
        """ Disposes of the contents of an editor.
        """
        disconnect(self._facename, wx.EVT_CHOICE)
        disconnect(self._point_size, wx.EVT_CHOICE)
        if self.factory.show_style:
            disconnect(self._style, wx.EVT_CHOICE)
        if self.factory.show_weight:
            disconnect(self._weight, wx.EVT_CHOICE)

        super(CustomFontEditor, self).dispose()

    #-------------------------------------------------------------------------
    #  Handles the user modifying one of the font components:
    #-------------------------------------------------------------------------

    def update_object_parts(self, event):
        """ Handles the user modifying one of the font components.
        """
        point_size = int(self._point_size.GetStringSelection())
        facename = self._facename.GetStringSelection()

        style = wx.FONTSTYLE_NORMAL
        if self.factory.show_style:
            style += self._style.GetCurrentSelection()

        weight = wx.FONTWEIGHT_NORMAL
        if self.factory.show_weight:
            weight += self._weight.GetCurrentSelection()

        font = wx.Font(point_size, wx.DEFAULT, style, weight,
                       faceName=facename)
        self.value = self.factory.from_wx_font(font)

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        font = self.factory.to_wx_font(self)

        try:
            self._facename.SetStringSelection(font.GetFaceName())
        except:
            self._facename.SetSelection(0)

        try:
            self._point_size.SetStringSelection(str(font.GetPointSize()))
        except:
            self._point_size.SetSelection(0)

    #-------------------------------------------------------------------------
    #  Returns the text representation of a specified font value:
    #-------------------------------------------------------------------------

    def string_value(self, font):
        """ Returns the text representation of a specified font value.
        """
        return self.factory.str_font(font)


#-------------------------------------------------------------------------
#  'TextFontEditor' class:
#-------------------------------------------------------------------------

class TextFontEditor(BaseTextEditor):
    """ Text style of font editor, which displays an editable text field
        containing a text representation of the font value (using that font if
        possible).
    """

    #-------------------------------------------------------------------------
    #  Handles the user changing the contents of the edit control:
    #-------------------------------------------------------------------------

    def update_object(self, event):
        """ Handles the user changing the contents of the edit control.
        """
        self.value = self.control.GetValue()

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes external to the
            editor.
        """
        super(TextFontEditor, self).update_editor()
        set_font(self)

    #-------------------------------------------------------------------------
    #  Returns the text representation of a specified font value:
    #-------------------------------------------------------------------------

    def string_value(self, font):
        """ Returns the text representation of a specified font value.
        """
        return self.factory.str_font(font)

#-------------------------------------------------------------------------
#  'ReadonlyFontEditor' class:
#-------------------------------------------------------------------------


class ReadonlyFontEditor(BaseReadonlyEditor):
    """ Read-only style of font editor, which displays a read-only text field
        containing a text representation of the font value (using that font if
        possible).
    """

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes external to the
            editor.
        """
        super(ReadonlyFontEditor, self).update_editor()
        set_font(self)

    #-------------------------------------------------------------------------
    #  Returns the text representation of a specified font value:
    #-------------------------------------------------------------------------

    def string_value(self, font):
        """ Returns the text representation of a specified font value.
        """
        return self.factory.str_font(font)

#-------------------------------------------------------------------------
#  Set the editor control's font to match a specified font:
#-------------------------------------------------------------------------


def set_font(editor):
    """ Sets the editor control's font to match a specified font.
    """
    font = editor.factory.to_wx_font(editor)
    font.SetPointSize(min(10, font.GetPointSize()))
    editor.control.SetFont(font)

#-------------------------------------------------------------------------
#  'FontEnumerator' class:
#-------------------------------------------------------------------------


class FontEnumerator(wx.FontEnumerator):
    """ An enumeration of fonts.
    """
    #-------------------------------------------------------------------------
    #  Returns a list of all available font facenames:
    #-------------------------------------------------------------------------

    def facenames(self):
        """ Returns a list of all available font facenames.
        """
        self._facenames = []
        self.EnumerateFacenames()
        return self._facenames

    #-------------------------------------------------------------------------
    #  Adds a facename to the list of facenames:
    #-------------------------------------------------------------------------

    def OnFacename(self, facename):
        """ Adds a facename to the list of facenames.
        """
        self._facenames.append(facename)
        return True

# Define the names SimpleEditor, CustomEditor, TextEditor and ReadonlyEditor
# which are looked up by the editor factory for the font editor.
SimpleEditor = SimpleFontEditor
CustomEditor = CustomFontEditor
TextEditor = TextFontEditor
ReadonlyEditor = ReadonlyFontEditor

### EOF #######################################################################



#------------------------------------------------------------------------------
#
#  Copyright (c) 2008, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   01/27/2006
#
#------------------------------------------------------------------------------

""" Defines the code editor factory for all traits toolkit backends,
useful for tools such as debuggers.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from traits.api import Instance, Str, Color, Enum, Bool

from ..editor_factory import EditorFactory

#-------------------------------------------------------------------------
#  'ToolkitEditorFactory' class:
#-------------------------------------------------------------------------


class ToolkitEditorFactory(EditorFactory):
    """ Editor factory for code editors.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Object trait containing list of line numbers to mark (optional)
    mark_lines = Str

    # Background color for marking lines
    mark_color = Color(0xECE9D8)

    # Object trait containing the currently selected line (optional)
    selected_line = Str

    # Object trait containing the currently selected text (optional)
    selected_text = Str

    # Object trait containing the currently selected text start position
    # (optional)
    selected_start_pos = Str

    # Object trait containing the currently selected text end position
    # (optional)
    selected_end_pos = Str

    # Background color for selected lines
    selected_color = Color(0xA4FFFF)

    # Where should the search toolbar be placed?
    search = Enum('top', 'bottom', 'none')

    # Background color for lines that match the current search
    search_color = Color(0xFFFF94)

    # Current line
    line = Str

    # Current column
    column = Str

    # Should code folding be enabled?
    foldable = Bool(True)

    # Should line numbers be displayed in the margin?
    show_line_numbers = Bool(True)

    # Is user input set on every change?
    auto_set = Bool(True)

    # Should the editor auto-scroll when a new **selected_line** value is set?
    auto_scroll = Bool(True)

    # Optional key bindings associated with the editor
    key_bindings = Instance('traitsui.key_bindings.KeyBindings')

    # Calltip clicked event
    calltip_clicked = Str

    # The lexer to use. Default is 'python'; 'null' indicates no lexing.
    lexer = Str('python')

    # Object trait containing the list of line numbers to dim (optional)
    dim_lines = Str

    # Object trait to dim lines to. Can be of form #rrggbb or a color spec. If
    # not specified, dark grey is used.
    dim_color = Str

    # Object trait containing the list of line numbers to put squiggles under
    # (optional)
    squiggle_lines = Str

    # Object trait for the color of squiggles. If not specified, red is used.
    squiggle_color = Str


# Define the Code Editor class.
CodeEditor = ToolkitEditorFactory

### EOF #######################################################################



#-----------------------------------------------------------------------------
#
#  Copyright (c) 2006 by Enthought, Inc.
#  All rights reserved.
#
#  Author: Dave Peterson <dpeterson@enthought.com>
#
#-----------------------------------------------------------------------------

""" Tests the global registry functionality of the sweet_pickle framework.
"""

# Standard library imports.
import unittest

# Enthought library imports
import apptools.sweet_pickle as sweet_pickle
from apptools.sweet_pickle.global_registry import _clear_global_registry


##############################################################################
# class 'GlobalRegistryTestCase'
##############################################################################

class GlobalRegistryTestCase(unittest.TestCase):
    """ Tests the global registry functionality of the sweet_pickle framework.
    """

    ##########################################################################
    # 'TestCase' interface
    ##########################################################################

    ### public interface #####################################################

    def setUp(self):
        """ Creates the test fixture.

            Overridden here to ensure each test starts with an empty global
            registry.
        """
        # Clear the global registry
        _clear_global_registry()

        # Cache a reference to the new global registry
        self.registry = sweet_pickle.get_global_registry()


    ##########################################################################
    # 'GlobalRegistryTestCase' interface
    ##########################################################################

    ### public interface #####################################################

    def test_clearing(self):
        """ Validates that clearing the registry gives us a new registry.
        """
        _clear_global_registry()
        self.assertNotEqual(self.registry, sweet_pickle.get_global_registry())


    def test_registry_starts_empty(self):
        """ Validates that the registry is starting empty for each test.
        """
        self.assertEqual(0, len(self.registry.class_map))
        self.assertEqual(0, len(self.registry.state_functions))
        self.assertEqual(0, len(self.registry.version_attribute_map))
        self.assertEqual(0, len(self.registry._state_function_classes))


    def test_returns_singleton(self):
        """ Validates that the getter returns the same global registry
        """
        # Just try it a few times.
        self.assertEqual(self.registry, sweet_pickle.get_global_registry())
        self.assertEqual(self.registry, sweet_pickle.get_global_registry())
        self.assertEqual(self.registry, sweet_pickle.get_global_registry())


if __name__ == "__main__":
    unittest.main()


### EOF ######################################################################



from traits.api import HasTraits, provides

from omnivore.file_type.i_file_recognizer import IFileRecognizer, RecognizerBase
import imghdr


@provides(IFileRecognizer)
class ImageRecognizer(RecognizerBase):
    """Recognizer for common image formats
    
    """
    id = "image/common"

    before = "text/plain"

    mime_map = {
        'rgb': 'x-rgb',
        'pbm': 'x-portable-bitmap',
        'pgm': 'x-portable-graymap',
        'ppm': 'x-portable-pixmap',
        'rast': 'x-cmu-raster',
        'xbm': 'x-xbitmap,'
        }

    def identify(self, guess):
        name = imghdr.what("", h=guess.get_utf8())
        if name is None:
            return
        name = self.mime_map.get(name, name)
        return "image/%s" % name



"""Trivial Interfaces and Adaptation from PyProtocols.

This package used to be a subset of the files from Phillip J. Eby's PyProtocols
package. The package has been substituted by :mod:`traits.adaptation` as of
Traits 4.4.0.

Currently, the package contains deprecated aliases for backward compatibility,
and will be removed in Traits 5.0 .

"""



from pyface.api import ImageResource
from pyface.action.api import Group
from pyface.tasks.action.api import SToolBar, EditorAction
from traits.api import on_trait_change, Any


class MouseHandlerBaseAction(EditorAction):
    """Save a bit of boilerplate with a base class for toolbar mouse mode buttons
    
    Note that the traits for name, tooltip, and image must be repeated
    in subclasses because the trait initialization appears to reference
    the handler in the class that is named, not superclasses.  E.g.:
    handler.menu_item_name in this base class doesn't appear to look at the
    handler class attribute of subclasses.
    """
    # Traits
    handler = Any

    style = 'radio'

    enabled_name = ''

    def _name_default(self):
        return self.handler.menu_item_name

    def _tooltip_default(self):
        return self.handler.menu_item_tooltip

    def _image_default(self):
        return ImageResource(self.handler.icon)

    def perform(self, event):
        self.active_editor.mouse_mode_factory = self.handler
        self.active_editor.update_mouse_mode()

    @on_trait_change('active_editor.mouse_mode_factory')
    def _update_checked(self):
        if self.active_editor:
            self.checked = self.active_editor.mouse_mode_factory == self.handler


def get_toolbar_group(toolbar_name, mouse_handlers):
    """Create the toolbar groups with buttons in the order specified in the
    valid_mouse_modes dict.
    """
    actions = [MouseHandlerBaseAction(handler=mode, enabled_name=mode.editor_trait_for_enabled) for mode in mouse_handlers]
    return SToolBar(Group(*actions),
                    show_tool_names=False,
                    # image_size=(22,22),
                    id=toolbar_name)



# Enthought library imports.
from envisage.ui.tasks.api import PreferencesPane, TaskFactory
from apptools.preferences.api import PreferencesHelper
from traits.api import Bool, Dict, Enum, List, Str, Unicode
from traitsui.api import EnumEditor, HGroup, VGroup, Item, Label, \
    View


class ImageEditPreferences(PreferencesHelper):
    """ The preferences helper for the Framework application.
    """

    #### 'PreferencesHelper' interface ########################################

    # The path to the preference node that contains the preferences.
    preferences_path = 'omnivore.framework.image_edit'

    #### Preferences ##########################################################

    # Scale image to fit window when initially loaded
    fit_image_on_load = Bool


class ImageEditPreferencesPane(PreferencesPane):
    """ The preferences pane for the Framework application.
    """

    #### 'PreferencesPane' interface ##########################################

    # The factory to use for creating the preferences model object.
    model_factory = ImageEditPreferences

    category = Str('Editors')

    #### 'FrameworkPreferencesPane' interface ################################

    view = View(
        VGroup(HGroup(Item('fit_image_on_load'),
                      Label('Scale image to window on initial load'),
                      show_labels = False),
               label='Image editor'),
        resizable=True)



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/21/2004
#
#------------------------------------------------------------------------------

""" Defines the compound editor and the compound editor factory for the
    wxPython user interface toolkit.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

import wx

from traits.api \
    import Str

# FIXME: ToolkitEditorFactory is a proxy class defined here just for backward
# compatibility. The class has been moved to the
# traitsui.editors.compound_editor file.
from traitsui.editors.compound_editor \
    import ToolkitEditorFactory

from editor \
    import Editor

from helper \
    import TraitsUIPanel

#-------------------------------------------------------------------------
#  'CompoundEditor' class:
#-------------------------------------------------------------------------


class CompoundEditor(Editor):
    """ Editor for compound traits, which displays editors for each of the
    combined traits, in the appropriate style.
    """
    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # The kind of editor to create for each list item
    kind = Str

    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        # Create a panel to hold all of the component trait editors:
        self.control = panel = TraitsUIPanel(parent, -1)
        sizer = wx.BoxSizer(wx.VERTICAL)

        # Add all of the component trait editors:
        self._editors = editors = []
        for factory in self.factory.editors:
            editor = getattr(
                factory,
                self.kind)(
                self.ui,
                self.object,
                self.name,
                self.description,
                panel)
            editor.prepare(panel)
            sizer.Add(editor.control, 1,
                      wx.TOP | wx.BOTTOM | editor.layout_style, 3)
            editors.append(editor)

        # Set-up the layout:
        panel.SetSizerAndFit(sizer)

        self.set_tooltip()

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        pass

    #-------------------------------------------------------------------------
    #  Disposes of the contents of an editor:
    #-------------------------------------------------------------------------

    def dispose(self):
        """ Disposes of the contents of an editor.
        """
        for editor in self._editors:
            editor.dispose()

        super(CompoundEditor, self).dispose()

#-------------------------------------------------------------------------
#  'SimpleEditor' class:
#-------------------------------------------------------------------------


class SimpleEditor(CompoundEditor):

    # The kind of editor to create for each list item. This value overrides
    # the default.
    kind = 'simple_editor'

#-------------------------------------------------------------------------
#  'CustomEditor' class:
#-------------------------------------------------------------------------


class CustomEditor(CompoundEditor):

    # The kind of editor to create for each list item. This value overrides
    # the default.

    kind = 'custom_editor'

#-- EOF ----------------------------------------------------------------------



#-------------------------------------------------------------------------
#
#  Copyright (c) 2008, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/21/2004
#
#-------------------------------------------------------------------------

""" Defines a text editor which displays a text field and maintains a history
    of previously entered values.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from traits.api import Int, Bool

from ..basic_editor_factory import BasicEditorFactory

from ..toolkit import toolkit_object

# Define callable which returns the 'klass' value (i.e. the editor to use in
# the EditorFactory.


def history_editor(*args, **traits):
    return toolkit_object('history_editor:_HistoryEditor')(*args, **traits)

#-------------------------------------------------------------------------
#  Create the editor factory object:
#-------------------------------------------------------------------------


class ToolkitEditorFactory(BasicEditorFactory):

    # The number of entries in the history:
    entries = Int(10)

    # Should each keystroke update the value (or only the enter key, tab,
    # etc.)?
    auto_set = Bool(False)


HistoryEditor = ToolkitEditorFactory(klass=history_editor)
# EOF #########################################################################
